"""
Functions for Google Sheets
"""

from typing import Optional, Union
import logging
import os

import pandas as pd
import requests

from google.oauth2.credentials import Credentials
from google.oauth2 import service_account
from google.auth.exceptions import RefreshError
from gspread_dataframe import set_with_dataframe
import gspread

from . import errors

LOGGER = logging.getLogger(__name__)


class MegatonGS(object):
    """Google Sheets client
    """
    required_scopes = [
        'https://www.googleapis.com/auth/spreadsheets',
        'https://www.googleapis.com/auth/drive',
    ]

    def __init__(self, credentials: Credentials, url: Optional[str] = None, timeout: Optional[float] = None):
        """constructor"""
        self.credentials = credentials
        self._client: gspread.client.Client = None
        self._driver: gspread.spreadsheet.Spreadsheet = None
        self.sheet = self.Sheet(self)
        self.timeout = self._resolve_timeout(timeout)

        self._authorize()
        if url:
            self.open(url)

    def _authorize(self):
        """Validate credentials given and build client"""
        if not isinstance(self.credentials, (Credentials, service_account.Credentials)):
            self.credentials = None
            raise errors.BadCredentialFormat
        elif self.credentials.scopes:
            if not set(self.required_scopes) <= set(self.credentials.scopes):
                self.credentials = None
                raise errors.BadCredentialScope(self.required_scopes)
        self._client = gspread.authorize(self.credentials)
        if self.timeout is not None:
            self._client.http_client.timeout = self.timeout

    def _resolve_timeout(self, timeout: Optional[float]) -> Optional[float]:
        if timeout is None:
            env = os.getenv("MEGATON_GS_TIMEOUT")
            if env is not None:
                try:
                    timeout = float(env)
                except ValueError:
                    timeout = 180.0
            else:
                timeout = 180.0
        if timeout is not None and timeout <= 0:
            return None
        return timeout

    @property
    def sheets(self):
        return [s.title for s in self._driver.worksheets()] if self._driver else []

    @property
    def title(self):
        return self._driver.title if self._driver else None

    @property
    def url(self):
        return self._driver.url if self._driver else None

    def open(self, url: str, sheet: str = None):
        """Get or create an api service

        Args:
            url (str): URL of the Google Sheets to open.
            sheet (str): Sheet name to open. optional

        Returns:
            title (str): the title of the Google Sheets opened
        """

        try:
            self._driver = self._client.open_by_url(url)
        except gspread.exceptions.NoValidUrlKeyFound:
            raise errors.BadUrlFormat
        except PermissionError as exc:
            raise errors.BadPermission from exc
        except requests.exceptions.Timeout as exc:
            raise errors.Timeout from exc
        except requests.exceptions.RequestException as exc:
            raise errors.RequestError(str(exc)) from exc

        try:
            title = self._driver.title
        except RefreshError:
            raise errors.BadCredentialScope
        except gspread.exceptions.APIError as e:
            if 'disabled' in str(e):
                raise errors.ApiDisabled
            elif 'PERMISSION_DENIED' in str(e):
                raise errors.BadPermission
            elif 'NOT_FOUND' in str(e):
                raise errors.UrlNotFound
            else:
                raise

        if sheet:
            self.sheet.select(sheet)

        return title

    class Sheet(object):
        def __init__(self, parent):
            """constructor"""
            self.parent = parent
            self._driver: gspread.worksheet.Worksheet = None
            self.cell = self.Cell(self)

        def _refresh(self):
            """Rebuild the Gspread client"""
            # self._driver = self.parent._driver.worksheet(self.name)
            self.select(self.name)

        def clear(self):
            """Blank all the cells on the sheet"""
            self._driver.clear()

        def create(self, name: str):
            if not self.parent._client:
                LOGGER.warn("Open URL first.")
                return
            self.parent._driver.add_worksheet(
                title=name, rows=100, cols=20)
            self.select(name)

        def delete(self, name: str):
            if not self.parent._client:
                LOGGER.error("Open URL first.")
                return
            try:
                ws = self.parent._driver.worksheet(name)
            except gspread.exceptions.WorksheetNotFound:
                raise errors.SheetNotFound
            self.parent._driver.del_worksheet(ws)
            if self._driver and self._driver.title == name:
                self._driver = None

        def select(self, name: str):
            if not self.parent._client:
                LOGGER.error("Open URL first.")
                return
            try:
                self._driver = self.parent._driver.worksheet(name)
            except gspread.exceptions.WorksheetNotFound:
                raise errors.SheetNotFound
            except gspread.exceptions.APIError as e:
                if 'disabled' in str(e):
                    raise errors.ApiDisabled
                elif 'PERMISSION_DENIED' in str(e):
                    raise errors.BadPermission
            return self.name

        @property
        def id(self):
            """Sheet ID"""
            return self._driver.id if self._driver else None

        @property
        def name(self):
            """Sheet Name"""
            return self._driver.title if self._driver else None

        @property
        def last_row(self):
            """looks for the last row based on values appearing in all columns
            """
            cols = self._driver.range(1, 1, self._driver.row_count, self._driver.col_count)
            last = [cell.row for cell in cols if cell.value]
            return max(last) if last else 0

        @property
        def next_available_row(self):
            """looks for the first empty row based on values appearing in all columns
            """
            return self.last_row + 1

        @property
        def data(self):
            """Returns a list of dictionaries, all of them having the contents of
                    the spreadsheet with the head row as keys and each of these
                    dictionaries holding the contents of subsequent rows of cells as
                    values.
            """
            if not self._driver:
                LOGGER.error("Please select a sheet first.")
                return
            data = self._driver.get_all_records()
            return data

        def auto_resize(self, cols: list):
            """Auto resize columns to fit text"""
            sheet_id = self.id
            _requests = []
            for i in cols:
                dim = {
                    "autoResizeDimensions": {
                        "dimensions": {
                            "sheetId": sheet_id,
                            "dimension": "COLUMNS",
                            "startIndex": i - 1,
                            "endIndex": i
                        }
                    }
                }
                _requests.append(dim)
            self.parent._driver.batch_update({'requests': _requests})

        def resize(self, col: int, width: int):
            """Resize columns"""
            sheet_id = self.id
            _requests = []
            for i in [col]:
                dim = {
                    "updateDimensionProperties": {
                        "range": {
                            "sheetId": sheet_id,
                            "dimension": "COLUMNS",
                            "startIndex": i - 1,
                            "endIndex": i
                        },
                        "properties": {
                            "pixelSize": width
                        },
                        "fields": "pixelSize"
                    }
                }
                _requests.append(dim)
            self.parent._driver.batch_update({'requests': _requests})

        def freeze(self, rows: Optional[int] = None, cols: Optional[int] = None):
            """Freeze rows and/or columns on the worksheet"""
            self._driver.freeze(rows=rows, cols=cols)

        def save_data(self, df: pd.DataFrame, mode: str = 'a', row: int = 1, include_index: bool = False):
            """Save the dataframe to the sheet"""
            if not len(df):
                LOGGER.info("no data to write.")
                return
            elif not self._driver:
                LOGGER.warn("Please select a sheet first.")
                return
            elif mode == 'w':
                try:
                    self.clear()
                except gspread.exceptions.APIError as e:
                    if 'disabled' in str(e):
                        raise errors.ApiDisabled
                    elif 'PERMISSION_DENIED' in str(e):
                        raise errors.BadPermission

                set_with_dataframe(
                    self._driver,
                    df,
                    include_index=include_index,
                    include_column_header=True,
                    row=row,
                    resize=True
                )
                return True
            elif mode == 'a':
                next_row = self.next_available_row
                current_row = self._driver.row_count
                new_rows = df.shape[0]
                if current_row < next_row + new_rows - 1:
                    LOGGER.debug(f"adding {next_row + new_rows - current_row - 1} rows")
                    self._driver.add_rows(next_row + new_rows - current_row)
                    self._refresh()

                set_with_dataframe(
                    self._driver,
                    df,
                    include_index=include_index,
                    include_column_header=False,
                    row=next_row,
                    resize=False
                )
                return True

        def overwrite_data(self, df: pd.DataFrame, include_index: bool = False):
            """Clear the sheet and save the dataframe"""
            return self.save_data(df, mode='w', include_index=include_index)

        def overwrite_data_from_row(self, df: pd.DataFrame, row: int, include_index: bool = False):
            """Preserve rows above ``row`` and overwrite data from ``row`` onward."""
            if row <= 1:
                return self.overwrite_data(df, include_index=include_index)

            if not len(df):
                LOGGER.info("no data to write.")
                return
            if not self._driver:
                LOGGER.warn("Please select a sheet first.")
                return

            try:
                # Keep rows above `row` and clear old payload from target row onward.
                self._driver.batch_clear([f"{row}:{self._driver.row_count}"])
            except gspread.exceptions.APIError as e:
                if 'disabled' in str(e):
                    raise errors.ApiDisabled
                elif 'PERMISSION_DENIED' in str(e):
                    raise errors.BadPermission
                raise

            header_rows = 1
            required_last_row = row + len(df) + header_rows - 1
            if self._driver.row_count < required_last_row:
                self._driver.add_rows(required_last_row - self._driver.row_count)
                self._refresh()

            required_cols = len(df.columns) + (1 if include_index else 0)
            if self._driver.col_count < required_cols:
                self._driver.add_cols(required_cols - self._driver.col_count)
                self._refresh()

            set_with_dataframe(
                self._driver,
                df,
                include_index=include_index,
                include_column_header=True,
                row=row,
                resize=False,
            )
            return True

        class Cell(object):
            def __init__(self, parent):
                self.parent = parent
                self.address: str = ''
                # self.data = None

            @property
            def data(self):
                if self.address:
                    return self.parent._driver.acell(self.address).value

            @data.setter
            def data(self, value):
                if self.address:
                    self.parent._driver.update(self.address, value)

            def select(self, row: Union[int, str], col: Optional[int] = None):
                if not self.parent._driver:
                    LOGGER.error("Please select a sheet first.")
                    return

                self.address = gspread.utils.rowcol_to_a1(row, col) if col else row
                return self.data



"""
Common Functions
"""

from urllib.parse import unquote
import re
from datetime import datetime, timedelta

import pandas as pd
import pytz
from dateutil.relativedelta import relativedelta

DEFAULT_TIMEZONE = "UTC"
DEFAULT_FILTER_OPERATORS = ("==", "!=", "=@", "!@", "=~", "!~", ">", ">=", "<", "<=")


def parse_filter_conditions(
    conditions: str,
    allowed_ops: tuple | list | None = None,
    *,
    error_class=ValueError,
    allowed_ops_label: str | None = None,
):
    """Parse legacy filter string into a list of conditions.

    Returns a list of dicts: {"field": str, "operator": str, "value": str}.
    """
    if not conditions:
        return []

    ops = tuple(allowed_ops) if allowed_ops is not None else DEFAULT_FILTER_OPERATORS
    op_pattern = "|".join(re.escape(op) for op in sorted(ops, key=len, reverse=True))
    # Field names may include GA4 custom prefixes such as "customEvent:..."
    # and "customUser:..." in addition to legacy symbols.
    pattern = re.compile(rf"^([\w_\+\/\(\)\.\-: ]+)({op_pattern})(.+)$")
    allowed = allowed_ops_label or " ".join(ops)

    parsed = []
    for raw in conditions.split(";"):
        cond = raw.strip()
        if not cond:
            continue
        match = pattern.search(cond)
        if not match:
            raise error_class(
                f"Invalid filter: '{cond}'. Expected <dimension><op><expression>. "
                f"Allowed operators: {allowed}"
            )
        field, operator, value = match.groups()
        parsed.append(
            {
                "field": field.strip(),
                "operator": operator,
                "value": value.strip(),
            }
        )
    return parsed


def is_integer(n):
    """Determines the provided string is an integer number."""
    try:
        float(n)
    except ValueError:
        return False
    else:
        return float(n).is_integer()


def extract_integer_from_string(s):
    """Extracts integer from string provided."""
    m = re.search(r'(\d+)', s)
    if m:
        return int(m.group(1))


def change_column_type(df: pd.DataFrame, to_date=None, to_datetime=None):
    """Changes column type in dataframe from str to date or datetime."""
    if not to_date:
        to_date = ['date', 'firstSessionDate']
    if not to_datetime:
        to_datetime = ['dateHour', 'dateHourMinute']

    for col in df.columns:
        if col in to_date:
            df[col] = pd.to_datetime(df[col], errors='coerce').dt.date
        if col in to_datetime:
            df[col] = pd.to_datetime(df[col], errors='coerce')

    return df


def replace_columns(df: pd.DataFrame, rules: list):
    """Converts dataframe columns using regex.

    Args
        df: dataframe to be converted
        rules: list of tuple (column name, regex, to)
    """
    for r in rules:
        col, rule, to = r
        try:
            df[col] = df[col].replace(rule, to, regex=True)
        except KeyError as e:
            print(e)
            pass


def prep_df(df, delete_columns: list = None, type_columns: dict = None, rename_columns: dict = None):
    """Processes dataframe

    Args
        delete_columns:
            list of column name to be deleted
        type_columns:
            dict of column name -> data type
            ex. {'pageviews': 'int32'}
        rename_columns:
            dict of column name -> new column name
    Returns
        processed dataframe
    """
    if len(df) > 0:
        if delete_columns:
            # delete
            df.drop(delete_columns, axis=1, inplace=True)
        if type_columns:
            # change type
            df = df.astype(type_columns)
        if rename_columns:
            # rename
            df.columns = df.columns.to_series().replace(rename_columns, regex=True)
    return df


def get_date_range(start_date: str, end_date: str, format_: str = '%Y-%m-%d'):
    """Converts date range to a list of each date in the range."""
    date_range = pd.date_range(start_date, end_date)
    return [d.strftime(format_) for d in date_range]


def get_chunked_list(original_list: list, chunk_size: int = 100):
    """Splits a list into chunks."""
    chunked_list = []
    for i in range(0, len(original_list), chunk_size):
        chunked_list.append(original_list[i:i + chunk_size])
    return chunked_list


def get_past_date(n_days=None, n_months=None, timezone=DEFAULT_TIMEZONE, return_date_obj=False):
    """Return a date string (YYYY-MM-DD) N days or months ago.

    Args:
        n_days: Number of days to subtract from today.
        n_months: Number of months to subtract (returns first day of that month).
        timezone: Olson timezone name used for "today".
        return_date_obj: When True, return ``datetime.date`` instead of string.

    Raises:
        ValueError: When both ``n_days`` and ``n_months`` are provided.
    """
    tz = pytz.timezone(timezone)
    now = datetime.now(tz)

    if n_days is None and n_months is None:
        result = now
    elif n_days is not None and n_months is not None:
        raise ValueError("Specify either 'n_days' or 'n_months', not both.")
    elif n_days is not None:
        result = now - timedelta(days=n_days)
    else:
        start_of_month = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        result = start_of_month - relativedelta(months=n_months)

    return result.date() if return_date_obj else result.strftime("%Y-%m-%d")


def get_clean_url(url: str, params_to_keep: list = None):
    """Remove parameters from URL Query String"""

    if not params_to_keep:
        params_to_keep = []

    if "?" in url:
        base_url, arglist = url.split("?", 1)
        args = arglist.split("&")
        new_args = []
        for arg in args:
            try:
                k, v = arg.split("=", 1)
            except ValueError:
                k = arg
            if k.lower() in params_to_keep:
                new_args.append(unquote(arg))
                # print(f"keeping {arg}")
            else:
                # print(f"deleting {arg}")
                pass
        if len(new_args) > 0:
            # if param remains
            return "?".join([base_url, "&".join(new_args)])
        else:
            # all params are removed
            return base_url
    else:
        return url



"""
SearchResult.apply_if() メソッドのテスト
"""
import pandas as pd
import pytest
from megaton.start import SearchResult
from types import SimpleNamespace


def test_apply_if_with_true_condition():
    """条件が True の場合、メソッドが適用される"""
    df = pd.DataFrame({
        'query': ['test1', 'test2'],
        'clicks': [5, 10],
        'impressions': [100, 200],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # 条件が True なので filter_clicks が適用される
    filtered = result.apply_if(True, 'filter_clicks', min=7)
    
    assert len(filtered.df) == 1
    assert filtered.df.iloc[0]['clicks'] == 10


def test_apply_if_with_false_condition():
    """条件が False の場合、メソッドが適用されない"""
    df = pd.DataFrame({
        'query': ['test1', 'test2'],
        'clicks': [5, 10],
        'impressions': [100, 200],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # 条件が False なのでフィルタは適用されない
    filtered = result.apply_if(False, 'filter_clicks', min=7)
    
    assert len(filtered.df) == 2
    assert list(filtered.df['clicks']) == [5, 10]


def test_apply_if_with_callable_condition():
    """callable な条件を使用できる"""
    df = pd.DataFrame({
        'query': ['test1', 'test2', 'test3'],
        'clicks': [5, 10, 15],
        'impressions': [100, 200, 300],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # データ量に応じた条件分岐
    filtered = result.apply_if(
        lambda sr: len(sr.df) > 2,  # 3行あるので True
        'filter_clicks',
        min=7
    )
    
    assert len(filtered.df) == 2  # clicks >= 7 のみ残る
    assert list(filtered.df['clicks']) == [10, 15]


def test_apply_if_callable_condition_false():
    """callable な条件が False の場合"""
    df = pd.DataFrame({
        'query': ['test1', 'test2'],
        'clicks': [5, 10],
        'impressions': [100, 200],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # 条件が False なのでフィルタは適用されない
    filtered = result.apply_if(
        lambda sr: len(sr.df) > 5,  # 2行しかないので False
        'filter_clicks',
        min=7
    )
    
    assert len(filtered.df) == 2
    assert list(filtered.df['clicks']) == [5, 10]


def test_apply_if_chaining():
    """メソッドチェーンで複数の apply_if を使用できる"""
    df = pd.DataFrame({
        'query': ['test1', 'test2', 'test3', 'test4'],
        'clicks': [1, 5, 10, 15],
        'impressions': [50, 100, 200, 300],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # 条件付きで複数のフィルタを適用
    filtered = (
        result
        .apply_if(True, 'filter_clicks', min=5)     # clicks >= 5
        .apply_if(True, 'filter_impressions', min=150)  # impressions >= 150
        .apply_if(False, 'filter_clicks', max=5)   # これは適用されない
    )
    
    assert len(filtered.df) == 2  # clicks >= 5 AND impressions >= 150
    assert list(filtered.df['clicks']) == [10, 15]


def test_apply_if_with_variable_condition():
    """実際のユースケース：TARGET_MONTHS_AGO 風の条件"""
    df = pd.DataFrame({
        'query': ['test1', 'test2', 'test3'],
        'clicks': [1, 5, 10],
        'impressions': [50, 100, 200],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    TARGET_MONTHS_AGO = 0  # 最新月
    
    # 最新月はフィルタなし
    filtered_latest = (
        result
        .apply_if(TARGET_MONTHS_AGO > 0, 'filter_impressions', min=100)
        .apply_if(TARGET_MONTHS_AGO > 0, 'filter_clicks', min=5)
    )
    
    assert len(filtered_latest.df) == 3  # フィルタされない
    
    TARGET_MONTHS_AGO = 1  # 過去月
    
    # 過去月はフィルタあり
    filtered_past = (
        result
        .apply_if(TARGET_MONTHS_AGO > 0, 'filter_impressions', min=100)
        .apply_if(TARGET_MONTHS_AGO > 0, 'filter_clicks', min=5)
    )
    
    assert len(filtered_past.df) == 2  # impressions >= 100 AND clicks >= 5
    assert list(filtered_past.df['clicks']) == [5, 10]


def test_apply_if_nonexistent_method():
    """存在しないメソッド名の場合は AttributeError"""
    df = pd.DataFrame({'query': ['test1'], 'clicks': [5]})
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    with pytest.raises(AttributeError):
        result.apply_if(True, 'nonexistent_method')


def test_apply_if_preserves_dimensions():
    """apply_if は dimensions を保持する"""
    df = pd.DataFrame({
        'query': ['test1', 'test2'],
        'page': ['/a', '/b'],
        'clicks': [5, 10],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query', 'page'])
    
    filtered = result.apply_if(True, 'filter_clicks', min=7)
    
    assert filtered.dimensions == ['query', 'page']
    assert len(filtered.df) == 1


def test_apply_if_with_kwargs():
    """kwargs を正しく渡せる"""
    df = pd.DataFrame({
        'query': ['test1', 'test2', 'test3'],
        'clicks': [1, 5, 10],
        'impressions': [50, 100, 200],
    })
    parent = SimpleNamespace()
    result = SearchResult(df, parent, ['query'])
    
    # min と max の両方を指定
    filtered = result.apply_if(True, 'filter_clicks', min=3, max=8)
    
    assert len(filtered.df) == 1
    assert filtered.df.iloc[0]['clicks'] == 5



"""Tests for ReportResult class"""
import pytest
import pandas as pd
from megaton.start import ReportResult


def test_report_result_init():
    """ReportResult の初期化テスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-02'],
        'sessionSource': ['google', 'yahoo'],
        'sessions': [100, 50]
    })
    
    result = ReportResult(df)
    
    assert isinstance(result, ReportResult)
    assert len(result) == 2
    assert result.columns == ['date', 'sessionSource', 'sessions']
    assert result.dimensions == ['date', 'sessionSource']
    assert not result.empty


def test_report_result_init_with_dimensions():
    """明示的な dimensions 指定"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100]
    })
    
    result = ReportResult(df, dimensions=['date'])
    
    assert result.dimensions == ['date']


def test_report_result_auto_detect_dimensions():
    """指標列の自動検出"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessionSource': ['google'],
        'sessions': [100],
        'users': [80],
        'activeUsers': [60]
    })
    
    result = ReportResult(df)
    
    # 指標列は dimensions に含まれない
    assert 'sessions' not in result.dimensions
    assert 'users' not in result.dimensions
    assert 'activeUsers' not in result.dimensions
    # ディメンション列のみが含まれる
    assert 'date' in result.dimensions
    assert 'sessionSource' in result.dimensions


def test_report_result_numeric_dimensions_protected():
    """数値型ディメンション（month, yearMonth等）が dimensions に正しく含まれることを確認"""
    df = pd.DataFrame({
        'month': [202401, 202402],  # 数値型だが既知のディメンション
        'yearMonth': [202401, 202402],  # 数値型だが既知のディメンション
        'sessionSource': ['google', 'yahoo'],
        'sessions': [100, 50],
        'cv': [10, 5],  # カスタムメトリクス（数値型）
    })
    
    result = ReportResult(df)
    
    # 数値型でも既知のディメンションは dimensions に含まれる
    assert 'month' in result.dimensions
    assert 'yearMonth' in result.dimensions
    assert 'sessionSource' in result.dimensions
    # メトリクスは除外される
    assert 'sessions' not in result.dimensions
    assert 'cv' not in result.dimensions
    
    # fill() で正しく動作することを確認
    result_filled = result.fill()
    # month と yearMonth がディメンションとして扱われることを確認（直積が作られる）
    assert len(result_filled) >= len(result)


def test_report_result_backward_compatibility():
    """後方互換性のテスト（DataFrame的な操作）"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100]
    })
    
    result = ReportResult(df)
    
    # __getitem__
    assert (result['date'] == df['date']).all()
    assert (result['sessions'] == df['sessions']).all()
    
    # len
    assert len(result) == len(df)
    
    # .df property
    pd.testing.assert_frame_equal(result.df, df)


def test_report_result_normalize_no_aggregate():
    """normalize() は上書きのみで集約しない"""
    df = pd.DataFrame({
        'sessionSource': [' Google ', 'Yahoo'],
        'sessions': [100, 50],
    })
    result = ReportResult(df)
    normalized = result.normalize(
        dimension='sessionSource',
        by={r'google': 'Search'},
    )

    assert len(normalized) == 2
    assert normalized.df['sessionSource'].tolist() == ['Search', 'yahoo']


def test_report_result_categorize_adds_column():
    """categorize() はカテゴリ列を追加し集約しない"""
    df = pd.DataFrame({
        'sessionSource': ['google', 'unknown'],
        'sessions': [100, 50],
    })
    result = ReportResult(df)
    categorized = result.categorize(
        dimension='sessionSource',
        by={'google': 'Google'},
        default='Others',
    )

    assert 'sessionSource_category' in categorized.columns
    assert categorized.df['sessionSource_category'].tolist() == ['Google', 'Others']
    assert len(categorized) == 2


def test_report_result_classify_aggregates():
    """classify() は正規化して必ず集約する"""
    df = pd.DataFrame({
        'sessionSource': [' Google ', 'google'],
        'sessions': [100, 50],
    })
    result = ReportResult(df)
    classified = result.classify(
        dimension='sessionSource',
        by={r'google': 'Search'},
    )

    assert len(classified) == 1
    assert classified.df['sessionSource'].iloc[0] == 'Search'
    assert classified.df['sessions'].iloc[0] == 150


def test_report_result_group():
    """group() メソッドのテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01', '2024-02'],
        'sessionSource': ['google', 'yahoo', 'google'],
        'sessions': [100, 50, 150]
    })
    
    result = ReportResult(df)
    grouped = result.group(by='sessionSource', metrics=['sessions'])
    
    assert len(grouped) == 2
    assert grouped.dimensions == ['sessionSource']
    assert grouped.df[grouped.df['sessionSource'] == 'google']['sessions'].values[0] == 250


def test_report_result_group_multiple_dimensions():
    """group() 複数ディメンションのテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01', '2024-02'],
        'sessionSource': ['google', 'google', 'yahoo'],
        'sessions': [100, 50, 150]
    })
    
    result = ReportResult(df)
    grouped = result.group(by=['date', 'sessionSource'])
    
    # google + google で集計されるため 2 行
    assert len(grouped) == 2
    assert grouped.dimensions == ['date', 'sessionSource']
    # 2024-01 google の sessions は 100 + 50 = 150
    assert grouped.df[(grouped.df['date'] == '2024-01') & 
                      (grouped.df['sessionSource'] == 'google')]['sessions'].values[0] == 150


def test_report_result_group_auto_detect_metrics():
    """group() 指標自動検出のテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01'],
        'sessions': [100, 50],
        'users': [80, 40]
    })
    
    result = ReportResult(df)
    grouped = result.group(by='date')
    
    assert grouped.df['sessions'].values[0] == 150
    assert grouped.df['users'].values[0] == 120


def test_report_result_group_method():
    """group() 集計方法のテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01'],
        'sessions': [100, 50]
    })
    
    result = ReportResult(df)
    
    # sum
    grouped_sum = result.group(by='date', metrics=['sessions'], method='sum')
    assert grouped_sum.df['sessions'].values[0] == 150
    
    # mean
    grouped_mean = result.group(by='date', metrics=['sessions'], method='mean')
    assert grouped_mean.df['sessions'].values[0] == 75


def test_report_result_group_empty_dataframe():
    """group() 空DataFrameのテスト"""
    # 空DataFrameでmetrics明示指定
    df = pd.DataFrame(columns=['date', 'sessionSource', 'sessions', 'users'])
    result = ReportResult(df, dimensions=['date', 'sessionSource'])
    
    grouped = result.group(by=['date', 'sessionSource'], metrics=['sessions', 'users'])
    
    assert grouped.df.empty
    assert list(grouped.df.columns) == ['date', 'sessionSource', 'sessions', 'users']
    assert grouped.dimensions == ['date', 'sessionSource']


def test_report_result_group_empty_auto_metrics():
    """group() 空DataFrameで自動metrics検出のテスト"""
    df = pd.DataFrame(columns=['date', 'sessions'])
    result = ReportResult(df, dimensions=['date'])
    
    grouped = result.group(by='date')
    
    assert grouped.df.empty
    assert 'date' in grouped.df.columns
    assert grouped.dimensions == ['date']


def test_report_result_group_missing_metrics():
    """group() 指定したmetricsが存在しない場合のテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01'],
        'sessions': [100, 50]
    })
    
    result = ReportResult(df)
    
    # 存在しないmetric列を指定
    grouped = result.group(by='date', metrics=['sessions', 'nonexistent_metric'])
    
    # 存在する列のみ集計される
    assert 'sessions' in grouped.df.columns
    assert 'nonexistent_metric' not in grouped.df.columns
    assert grouped.df['sessions'].values[0] == 150


def test_report_result_group_empty_with_ghost_metrics():
    """group() 空DataFrameで存在しない列を指定した場合のテスト"""
    df = pd.DataFrame(columns=['date', 'sessions'])
    result = ReportResult(df, dimensions=['date'])
    
    # 存在しないmetric列を指定
    grouped = result.group(by='date', metrics=['sessions', 'ghost'])
    
    # 存在する列のみ含まれる
    assert grouped.df.empty
    assert 'sessions' in grouped.df.columns
    assert 'ghost' not in grouped.df.columns
    assert list(grouped.df.columns) == ['date', 'sessions']


def test_report_result_group_metrics_as_string():
    """group() metricsを文字列で指定した場合のテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01'],
        'sessions': [100, 50]
    })
    
    result = ReportResult(df)
    
    # metricsを文字列で指定（リストではなく）
    grouped = result.group(by='date', metrics='sessions')
    
    assert 'sessions' in grouped.df.columns
    assert grouped.df['sessions'].values[0] == 150
    assert grouped.dimensions == ['date']


def test_report_result_sort():
    """sort() メソッドのテスト"""
    df = pd.DataFrame({
        'sessionSource': ['yahoo', 'google', 'direct'],
        'sessions': [50, 100, 30]
    })
    
    result = ReportResult(df)
    sorted_result = result.sort(by='sessions', ascending=False)
    
    assert sorted_result.df['sessions'].values[0] == 100
    assert sorted_result.df['sessions'].values[1] == 50
    assert sorted_result.df['sessions'].values[2] == 30


def test_report_result_sort_multiple_columns():
    """sort() 複数列ソートのテスト"""
    df = pd.DataFrame({
        'date': ['2024-02', '2024-01', '2024-02'],
        'sessions': [100, 50, 150]
    })
    
    result = ReportResult(df)
    sorted_result = result.sort(by=['date', 'sessions'], ascending=[True, False])
    
    assert sorted_result.df['date'].values[0] == '2024-01'
    assert sorted_result.df['date'].values[1] == '2024-02'
    assert sorted_result.df['sessions'].values[1] == 150


def test_report_result_fill():
    """fill() メソッドのテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', None, '2024-02'],
        'sessionSource': ['google', 'yahoo', None],
        'sessions': [100, 50, 30]
    })
    
    result = ReportResult(df)
    filled = result.fill()
    
    assert filled.df['date'].isna().sum() == 0
    assert filled.df['sessionSource'].isna().sum() == 0
    assert filled.df['date'].values[1] == '(not set)'
    assert filled.df['sessionSource'].values[2] == '(not set)'


def test_report_result_fill_custom_value():
    """fill() カスタム値のテスト"""
    df = pd.DataFrame({
        'date': [None],
        'sessions': [100]
    })
    
    result = ReportResult(df)
    filled = result.fill(to='Unknown')
    
    assert filled.df['date'].values[0] == 'Unknown'


def test_report_result_fill_specific_dimensions():
    """fill() 特定ディメンションのテスト"""
    df = pd.DataFrame({
        'date': [None],
        'sessionSource': [None],
        'sessions': [100]
    })
    
    result = ReportResult(df, dimensions=['date', 'sessionSource'])
    filled = result.fill(dimensions=['date'])
    
    assert filled.df['date'].values[0] == '(not set)'
    assert pd.isna(filled.df['sessionSource'].values[0])


def test_report_result_to_int():
    """to_int() メソッドのテスト"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100.5],
        'users': [80.9]
    })
    
    result = ReportResult(df)
    int_result = result.to_int(metrics='sessions')
    
    assert int_result.df['sessions'].dtype == 'int64'
    assert int_result.df['sessions'].values[0] == 100
    assert int_result.df['users'].dtype == 'float64'


def test_report_result_to_int_multiple_metrics():
    """to_int() 複数指標のテスト"""
    df = pd.DataFrame({
        'sessions': [100.5],
        'users': [80.9]
    })
    
    result = ReportResult(df)
    int_result = result.to_int(metrics=['sessions', 'users'])
    
    assert int_result.df['sessions'].dtype == 'int64'
    assert int_result.df['users'].dtype == 'int64'


def test_report_result_to_int_auto_infer():
    """to_int() metrics=None で数値列の自動推論"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100.5],
        'users': [80.9],
        'channel': ['Organic Search']
    })
    
    result = ReportResult(df)
    int_result = result.to_int()  # metrics=None
    
    # 数値列は整数型に変換される
    assert int_result.df['sessions'].dtype == 'int64'
    assert int_result.df['users'].dtype == 'int64'
    assert int_result.df['sessions'].values[0] == 100
    assert int_result.df['users'].values[0] == 80
    
    # 非数値列は変換されない
    assert int_result.df['date'].dtype == 'object'
    assert int_result.df['channel'].dtype == 'object'


def test_report_result_to_int_with_fill_value():
    """to_int() でfill_valueのキーワード専用引数"""
    df = pd.DataFrame({
        'sessions': [100.5, None],
        'users': [80.9, None]
    })
    
    result = ReportResult(df)
    int_result = result.to_int(metrics=['sessions'], fill_value=99)
    
    assert int_result.df['sessions'].values[1] == 99
    assert pd.isna(int_result.df['users'].values[1])  # 変換されていない


def test_report_result_to_int_with_nan():
    """to_int() 欠損値のテスト"""
    df = pd.DataFrame({
        'sessions': [100.5, None, 50.2]
    })
    
    result = ReportResult(df)
    int_result = result.to_int(metrics='sessions', fill_value=0)
    
    assert int_result.df['sessions'].values[1] == 0


def test_report_result_replace():
    """replace() メソッドのテスト（固定文字列）"""
    df = pd.DataFrame({
        'sessionSource': ['google', 'yahoo', 'direct'],
        'sessions': [100, 50, 30]
    })
    
    result = ReportResult(df)
    replaced = result.replace(
        dimension='sessionSource',
        by={'google': 'Google', 'yahoo': 'Yahoo!'},
        regex=False  # 固定文字列として扱う
    )
    
    assert replaced.df['sessionSource'].values[0] == 'Google'
    assert replaced.df['sessionSource'].values[1] == 'Yahoo!'
    assert replaced.df['sessionSource'].values[2] == 'direct'


def test_report_result_replace_regex():
    """replace() メソッドのテスト（正規表現、default）"""
    df = pd.DataFrame({
        'campaign': ['test(123)', 'hello(world)', 'plain'],
        'sessions': [100, 50, 30]
    })
    
    result = ReportResult(df)
    
    # 正規表現で括弧内を削除（regex=True がデフォルト）
    replaced = result.replace(
        dimension='campaign',
        by={r'\([^)]*\)': ''}
    )
    
    assert replaced.df['campaign'].values[0] == 'test'
    assert replaced.df['campaign'].values[1] == 'hello'
    assert replaced.df['campaign'].values[2] == 'plain'


def test_report_result_replace_regex_multiple():
    """replace() メソッドで複数の正規表現パターンを適用"""
    df = pd.DataFrame({
        'campaign': ['Sale(2024)', 'New(Launch)', 'plain'],
        'sessions': [100, 50, 30]
    })
    
    result = ReportResult(df)
    
    # 括弧内を削除する正規表現
    replaced = result.replace(
        dimension='campaign',
        by={r'\(.*?\)': ''}
    )
    
    assert replaced.df['campaign'].values[0] == 'Sale'
    assert replaced.df['campaign'].values[1] == 'New'
    assert replaced.df['campaign'].values[2] == 'plain'


def test_report_result_clean_url():
    """clean_url() のテスト"""
    df = pd.DataFrame({
        'page': [
            'https://Example.com/Path/?a=1#top',
            '/deilab/?q=1',
        ],
        'sessions': [100, 50]
    })

    result = ReportResult(df)
    cleaned = result.clean_url('page')

    assert cleaned.df['page'].values[0] == 'https://example.com/path/'
    assert cleaned.df['page'].values[1] == '/deilab/'


def test_report_result_clean_url_missing_column_raises():
    """clean_url() 存在しない列はエラー"""
    df = pd.DataFrame({
        'query': ['a'],
        'sessions': [1],
    })

    result = ReportResult(df)
    with pytest.raises(ValueError):
        result.clean_url('page')


def test_report_result_method_chaining():
    """メソッドチェーンのテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01', None],
        'sessionSource': ['google', 'yahoo', 'direct'],
        'sessions': [100.5, 50.2, 30.8]
    })
    
    result = (ReportResult(df)
        .fill(to='(no date)', dimensions=['date'])
        .categorize(
            dimension='sessionSource',
            by={'.*google.*': 'Search', '.*yahoo.*': 'Search'},
            into='source_type',
        )
        .group(by=['date', 'source_type'], metrics=['sessions'])
        .to_int(metrics='sessions')
        .sort(by='sessions', ascending=False)
    )
    
    assert 'source_type' in result.columns
    assert result.df['sessions'].dtype == 'int64'
    assert result.df['sessions'].values[0] == result.df['sessions'].max()


def test_report_result_empty_dataframe():
    """空のDataFrameのテスト"""
    df = pd.DataFrame()
    result = ReportResult(df)
    
    assert result.empty
    assert len(result) == 0


def test_report_result_repr():
    """__repr__() のテスト"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-02'],
        'sessions': [100, 50]
    })
    
    result = ReportResult(df)
    repr_str = repr(result)
    
    assert 'ReportResult' in repr_str
    assert '2 rows' in repr_str
    assert '2 columns' in repr_str


def test_report_result_classify_missing_column():
    """classify() 存在しない列のエラーテスト"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100]
    })
    
    result = ReportResult(df)
    
    with pytest.raises(ValueError, match="Column 'nonexistent' not found"):
        result.classify(dimension='nonexistent', by={'x': 'y'})


def test_report_result_replace_missing_column():
    """replace() 存在しない列のエラーテスト"""
    df = pd.DataFrame({
        'date': ['2024-01'],
        'sessions': [100]
    })
    
    result = ReportResult(df)
    
    with pytest.raises(ValueError, match="Column 'nonexistent' not found"):
        result.replace(dimension='nonexistent', by={'x': 'y'})


def test_report_result_custom_metrics_not_in_dimensions():
    """カスタムメトリクス（cv, ad_cost, totalPurchasers等）が dimensions に入らないことを確認"""
    df = pd.DataFrame({
        'date': ['2024-01', '2024-01'],
        'sessionSource': ['google', 'yahoo'],
        'sessions': [100, 50],
        'cv': [10, 5],  # カスタムメトリクス（数値型）
        'ad_cost': [1000.5, 500.25],  # カスタムメトリクス（数値型）
        'totalPurchasers': [8, 4],  # 標準メトリクスだがKNOWN_GA4_METRICSに追加済み
    })
    
    result = ReportResult(df)
    
    # カスタムメトリクスは dimensions に含まれない（数値型として自動検出）
    assert 'cv' not in result.dimensions
    assert 'ad_cost' not in result.dimensions
    assert 'totalPurchasers' not in result.dimensions
    # ディメンションのみが含まれる
    assert result.dimensions == ['date', 'sessionSource']
    
    # classify() で正しく集計されることを確認
    classified = result.classify(
        dimension='sessionSource',
        by={'google': 'Search', 'yahoo': 'Search'},
    )
    
    # 1行に集計される
    assert len(classified) == 1
    assert classified['sessionSource'].iloc[0] == 'Search'
    assert classified['sessions'].iloc[0] == 150  # 100 + 50
    assert classified['cv'].iloc[0] == 15  # 10 + 5
    assert classified['ad_cost'].iloc[0] == 1500.75  # 1000.5 + 500.25
    assert classified['totalPurchasers'].iloc[0] == 12  # 8 + 4



import importlib
import json
import os
import sys
import types

import logging
import pytest


# ---------------------------------------------------------------------------
# Provide lightweight stand-ins for Google libraries when they are unavailable.
# This keeps the auth helpers importable inside a minimal test environment.
# ---------------------------------------------------------------------------

def ensure_module(name: str):
    parts = name.split('.')
    module_name = ''
    parent = None
    for part in parts:
        module_name = f"{module_name}.{part}" if module_name else part
        module = sys.modules.get(module_name)
        if module is None:
            module = types.ModuleType(module_name)
            sys.modules[module_name] = module
            if parent is not None:
                setattr(parent, part, module)
        parent = module
    return module


def get_module(name: str):
    try:
        return importlib.import_module(name)
    except Exception:
        return ensure_module(name)


credentials_mod = get_module('google.oauth2.credentials')
if not hasattr(credentials_mod, 'Credentials'):
    class _StubCredentials:
        pass

    credentials_mod.Credentials = _StubCredentials

service_account_mod = get_module('google.oauth2.service_account')
if not hasattr(service_account_mod, 'Credentials'):
    class _StubServiceAccount:
        pass

    service_account_mod.Credentials = _StubServiceAccount

flow_mod = get_module('google_auth_oauthlib.flow')
if not hasattr(flow_mod, 'InstalledAppFlow'):
    class _StubFlow:
        pass

    flow_mod.InstalledAppFlow = _StubFlow

transport_mod = get_module('google.auth.transport.requests')
if not hasattr(transport_mod, 'Request'):
    transport_mod.Request = lambda: object()

exceptions_mod = get_module('google.auth.exceptions')
if not hasattr(exceptions_mod, 'RefreshError'):
    class _StubRefreshError(Exception):
        pass

    exceptions_mod.RefreshError = _StubRefreshError


from megaton import auth


def test_is_service_account_detects_service_account_json():
    service_account_json = json.dumps({"type": "service_account", "project_id": "demo"})
    assert auth._is_service_account(service_account_json)
    assert not auth._is_service_account("not-json")
    assert not auth._is_service_account(json.dumps({"type": "user"}))


def test_get_credential_type_identifies_known_types():
    web_config = {
        "web": {
            "auth_uri": "https://example.com/auth",
            "token_uri": "https://example.com/token",
            "client_id": "abc123",
        }
    }
    installed_config = {
        "installed": {
            "auth_uri": "https://example.com/auth",
            "token_uri": "https://example.com/token",
            "client_id": "xyz789",
        }
    }

    assert auth.get_credential_type({"type": "service_account"}) == "service_account"
    assert auth.get_credential_type(web_config) == "web"
    assert auth.get_credential_type(installed_config) == "installed"
    assert auth.get_credential_type({"installed": {}}) is None


def test_get_credential_type_from_file(tmp_path):
    sa_file = tmp_path / "service.json"
    sa_file.write_text(json.dumps({"type": "service_account"}))

    oauth_file = tmp_path / "oauth.json"
    oauth_file.write_text(
        json.dumps(
            {
                "installed": {
                    "auth_uri": "a",
                    "token_uri": "b",
                    "client_id": "c",
                }
            }
        )
    )

    assert auth.get_credential_type_from_file(sa_file) == "service_account"
    assert auth.get_credential_type_from_file(oauth_file) == "installed"


def test_get_json_files_from_dir_classifies_files(tmp_path):
    root = tmp_path / "creds"
    root.mkdir()
    sa_dir = root / "sa"
    sa_dir.mkdir()
    oauth_dir = root / "oauth"
    oauth_dir.mkdir()

    (sa_dir / "sa.json").write_text(json.dumps({"type": "service_account"}))
    (oauth_dir / "client.json").write_text(
        json.dumps(
            {
                "installed": {
                    "auth_uri": "a",
                    "token_uri": "b",
                    "client_id": "c",
                }
            }
        )
    )

    result = auth.get_json_files_from_dir(root)
    assert result["Service Account"]["sa.json"] == os.path.join(sa_dir, "sa.json")
    assert result["OAuth"]["client.json"] == os.path.join(oauth_dir, "client.json")


def test_get_credential_type_from_info_handles_variants():
    assert auth.get_credential_type_from_info({"type": "service_account"}) == "service_account"
    assert auth.get_credential_type_from_info({"installed": {}}) == "installed"
    assert auth.get_credential_type_from_info({"web": {}}) == "web"
    assert auth.get_credential_type_from_info({}) == "unknown"


def test_get_cache_path_uses_home_config_dir(tmp_path, monkeypatch):
    monkeypatch.setattr(auth.google_auth.os.path, "expanduser", lambda _: str(tmp_path))
    cache = auth.get_cache_path("client-secret.json")
    assert cache.endswith(".config/cache_client-secret.json")
    assert os.path.isdir(os.path.join(tmp_path, ".config"))

def test_save_credentials_writes_json(tmp_path, monkeypatch):
    cache_file = tmp_path / "cache.json"
    monkeypatch.setattr(auth.google_auth, "get_cache_path", lambda _: str(cache_file))

    class DummyCred:
        def __init__(self):
            self.data = {"token": "abc"}

        def to_json(self):
            return json.dumps(self.data)

    cred = DummyCred()
    returned = auth.save_credentials("client.json", cred)
    assert returned is cred
    assert cache_file.read_text() == json.dumps({"token": "abc"})


def test_load_credentials_returns_none_when_cache_missing(tmp_path, monkeypatch):
    cache_file = tmp_path / "missing.json"
    monkeypatch.setattr(auth.google_auth, "get_cache_path", lambda _: str(cache_file))
    assert auth.load_credentials("client.json", scopes=["scope"]) is None


def test_load_credentials_reads_from_cache(monkeypatch, tmp_path):
    cache_file = tmp_path / "cache.json"
    cache_file.write_text("{}")
    monkeypatch.setattr(auth.google_auth, "get_cache_path", lambda _: str(cache_file))

    class DummyCredentials:
        last_call = None

        @classmethod
        def from_authorized_user_file(cls, filename, scopes):
            cls.last_call = (filename, tuple(scopes))
            return cls()

    monkeypatch.setattr(auth.google_auth, "Credentials", DummyCredentials)
    creds = auth.load_credentials("client.json", scopes=["a", "b"])
    assert isinstance(creds, DummyCredentials)
    assert DummyCredentials.last_call == (str(cache_file), ("a", "b"))




def test_load_service_account_credentials_from_info_refreshes(monkeypatch):
    class DummyCred:
        def __init__(self, info):
            self.valid = False
            self.refreshed = False
            self.scopes = None
            self.info = info
            self.service_account_email = info.get('client_email')

        def refresh(self, request):
            self.valid = True
            self.refreshed = True

    class DummyFactory:
        last = None

        @classmethod
        def from_service_account_info(cls, info, scopes=None):
            cred = DummyCred(info)
            cred.scopes = tuple(scopes) if scopes is not None else None
            cls.last = cred
            return cred

    monkeypatch.setattr(auth.google_auth.service_account, 'Credentials', DummyFactory)
    monkeypatch.setattr(auth.google_auth.google.auth.transport.requests, 'Request', lambda: object())
    result = auth.load_service_account_credentials_from_info({'type': 'service_account', 'client_email': 'sa@example.com'}, ['scope-a'])
    assert result is DummyFactory.last
    assert result.valid
    assert result.refreshed
    assert result.scopes == ('scope-a',)
def test_load_service_account_credentials_from_info_refresh_failure(monkeypatch, caplog):
    class DummyCred:
        def __init__(self, info):
            self.valid = False
            self.service_account_email = info.get('client_email')

        def refresh(self, request):
            raise auth.google_auth.google.auth.exceptions.RefreshError('boom')

    class DummyFactory:
        @classmethod
        def from_service_account_info(cls, info, scopes=None):
            return DummyCred(info)

    monkeypatch.setattr(auth.google_auth.service_account, 'Credentials', DummyFactory)
    monkeypatch.setattr(auth.google_auth.google.auth.transport.requests, 'Request', lambda: object())
    with caplog.at_level(logging.ERROR):
        result = auth.load_service_account_credentials_from_info({'type': 'service_account', 'client_email': 'sa@example.com'}, ['scope'])
    assert result is None
    assert 'sa@example.com' in caplog.text
    assert '存在しない' in caplog.text


def test_load_service_account_credentials_from_file_refresh_failure(monkeypatch, tmp_path, caplog):
    sa_path = tmp_path / 'sa.json'
    sa_path.write_text('{}')

    class DummyCred:
        def __init__(self):
            self.valid = False
            self.service_account_email = 'sa@example.com'

        def refresh(self, request):
            raise auth.google_auth.google.auth.exceptions.RefreshError('boom')

    class DummyFactory:
        @classmethod
        def from_service_account_file(cls, path, scopes=None):
            return DummyCred()

    monkeypatch.setattr(auth.google_auth.service_account, 'Credentials', DummyFactory)
    monkeypatch.setattr(auth.google_auth.google.auth.transport.requests, 'Request', lambda: object())

    with caplog.at_level(logging.ERROR):
        result = auth.load_service_account_credentials_from_file(str(sa_path), ['scope'])
    assert result is None
    assert 'sa@example.com' in caplog.text
    assert '存在しない' in caplog.text

def test_delete_credentials_removes_file(tmp_path):
    cache_file = tmp_path / "cache.json"
    cache_file.write_text("data")
    assert cache_file.exists()
    auth.delete_credentials(str(cache_file))
    assert not cache_file.exists()



from types import SimpleNamespace

from google.api_core.exceptions import ServiceUnavailable

from megaton import ga4


class _FakeDataClient:
    def __init__(self, responses):
        self._responses = list(responses)
        self.calls = 0

    def run_report(self, request):
        self.calls += 1
        result = self._responses.pop(0)
        if isinstance(result, Exception):
            raise result
        return result


class _DummyParent:
    def __init__(self, data_client):
        self.data_client = data_client


def test_request_report_api_retries_service_unavailable_then_succeeds(monkeypatch):
    client = _FakeDataClient([
        ServiceUnavailable("temporary"),
        SimpleNamespace(row_count=7),
    ])
    report = ga4.MegatonGA4.Report(_DummyParent(client))
    report._parse_response = lambda response: ([["ok"]], ["date"], ["TYPE_INTEGER"]) if response else ([], [], [])

    waits = []
    monkeypatch.setattr("megaton.ga4.time.sleep", waits.append)

    data, total_rows, headers, types = report._request_report_api(
        0,
        SimpleNamespace(),
        max_retries=2,
        backoff_factor=1.5,
    )

    assert client.calls == 2
    assert waits == [1.5]
    assert total_rows == 7
    assert data == [["ok"]]
    assert headers == ["date"]
    assert types == ["TYPE_INTEGER"]


def test_request_report_api_returns_empty_after_retry_exhaustion(monkeypatch):
    client = _FakeDataClient([
        ServiceUnavailable("temporary-1"),
        ServiceUnavailable("temporary-2"),
        ServiceUnavailable("temporary-3"),
    ])
    report = ga4.MegatonGA4.Report(_DummyParent(client))

    waits = []
    monkeypatch.setattr("megaton.ga4.time.sleep", waits.append)

    data, total_rows, headers, types = report._request_report_api(
        0,
        SimpleNamespace(),
        max_retries=3,
        backoff_factor=2.0,
    )

    assert client.calls == 3
    assert waits == [2.0, 4.0]
    assert total_rows == 0
    assert data == []
    assert headers == []
    assert types == []



"""
Functions for widgets (forms)
"""

from typing import List, Tuple

_IMPORT_ERROR = None
try:
    import ipywidgets as _ipywidgets
except Exception as exc:
    _ipywidgets = None
    _IMPORT_ERROR = exc


class WidgetsUnavailableError(ImportError):
    """Raised when ipywidgets is required but unavailable."""


def require_widgets():
    global _ipywidgets, _IMPORT_ERROR
    if _ipywidgets is not None:
        return _ipywidgets
    try:
        import ipywidgets as _ipywidgets  # noqa: F811
    except Exception as exc:
        _IMPORT_ERROR = exc
        raise WidgetsUnavailableError(
            "ipywidgets is required for UI. Install ipywidgets or use headless=True."
        ) from exc
    return _ipywidgets


def dropdown_menu(label: str, default: str, option_list: List[Tuple[str, str]] = None, width: Tuple[str, str] = None):
    """Create a drop-down menu
    """
    widgets = require_widgets()
    Dropdown = widgets.Dropdown
    option_list = option_list or []
    description_width, menu_width = (None, None)
    if width:
        description_width, menu_width = width

    # set label
    options = [(default, '')] if default else []

    # add options
    if option_list:
        options.extend(option_list)

    style = {'description_width': description_width} if description_width else {}
    layout = {'width': menu_width} if menu_width else {}

    return Dropdown(description=f"{label}: ", options=options, style=style, layout=layout)


def create_blank_menu(name: str = None, default: str = None, width: Tuple[str, str] = None):
    """空のセレクトメニューを作る

    Args
        name: label
        default: default value for the menu
        width: Tuple of label_width and menu_width in str
            label_width
            menu_width: 'max-content', 'initial'
    """
    return dropdown_menu(label=name, default=default, width=width)


def menu_for_credentials(json_files: dict):
    widgets = require_widgets()
    Dropdown = widgets.Dropdown
    Layout = widgets.Layout
    options = [('GAにアクセスする認証情報を選択してください', '')]

    options.append(('OAuth', ''))
    for f, p in json_files['OAuth'].items():
        options.append((f'-   {f}', p))

    options.append(('Service Account', ''))
    for f, p in json_files['Service Account'].items():
        options.append((f'-   {f}', p))

    # return dropdown_menu('JSON', '選択して', options)
    return Dropdown(
        description="JSON: ",
        options=options,
        layout=Layout(width='80%')
    )


def html_text(value: str, placeholder: str, description: str):
    widgets = require_widgets()
    HTML = widgets.HTML
    return HTML(
        value=value,
        placeholder=placeholder,
        description=description
    )


def input_text(value: str, placeholder: str, description: str, disabled: bool = False, width: Tuple[str, str] = None):
    widgets = require_widgets()
    Text = widgets.Text
    description_width, menu_width = width
    style = {'description_width': description_width} if description_width else None
    layout = {'width': menu_width}

    return Text(
        value=value,
        placeholder=placeholder,
        description=description,
        style=style,
        layout=layout,
        disabled=disabled,
    )


def tab(ga: list):
    widgets = require_widgets()
    VBox = widgets.VBox
    return VBox(ga)


def tab_set(tabs: list, titles: list):
    widgets = require_widgets()
    Tab = widgets.Tab
    tab_widget = Tab(children=tabs)
    for i, title in enumerate(titles):
        tab_widget.set_title(i, title)
    return tab_widget


def Output(*args, **kwargs):
    widgets = require_widgets()
    return widgets.Output(*args, **kwargs)


__all__ = [
    "WidgetsUnavailableError",
    "require_widgets",
    "dropdown_menu",
    "create_blank_menu",
    "menu_for_credentials",
    "html_text",
    "input_text",
    "tab",
    "tab_set",
    "Output",
]



import pandas as pd
from megaton.start import Megaton


def test_filter_by_thresholds_basic():
    # Create Search instance without full Megaton initialization (avoid auth side-effects)
    search = Megaton.Search(object())

    df = pd.DataFrame(
        {
            "query": ["a", "b", "c", "d"],
            "impressions": [100, 5, 50, 200],
            "position": [1.2, 8.5, 3.0, 10.0],
            "pv": [10, 0, 2, 20],
            "cv": [1, 0, 0, 2],
        }
    )

    site = {"min_impressions": 50, "max_position": 5.0, "min_pv": 1, "min_cv": 1}

    filtered = search.filter_by_thresholds(df, site)
    # only rows meeting all: impressions>=50, position<=5.0, pv>=1, cv>=1
    assert list(filtered["query"]) == ["a"]


def test_filter_by_thresholds_missing_columns():
    search = Megaton.Search(object())

    df = pd.DataFrame({"query": ["x", "y"], "impressions": [10, 100], "position": [2, 3]})
    site = {"min_impressions": 50, "max_position": 5.0}
    filtered = search.filter_by_thresholds(df, site)
    assert list(filtered["query"]) == ["y"]


def test_filter_by_thresholds_clicks_zero_only():
    """Test that clicks_zero_only parameter filters only zero-click rows."""
    search = Megaton.Search(object())

    # Test data matching problem statement scenario
    df = pd.DataFrame(
        {
            "query": ["a", "b", "c"],
            "impressions": [100, 5, 50],
            "position": [1.0, 10.0, 3.0],
            "clicks": [1, 0, 0],
        }
    )

    site = {"min_impressions": 50, "max_position": 5}

    # Default behavior (clicks_zero_only=False): apply thresholds to all rows
    filtered_no_flag = search.filter_by_thresholds(df, site)
    # "a": impressions=100>=50 ✓, position=1.0<=5 ✓ → KEPT
    # "b": impressions=5<50 ✗ → EXCLUDED
    # "c": impressions=50>=50 ✓, position=3.0<=5 ✓ → KEPT
    assert sorted(list(filtered_no_flag["query"])) == ["a", "c"]

    # With clicks_zero_only=True: preserve rows with clicks>0, filter only clicks==0 rows
    filtered_zero_flag = search.filter_by_thresholds(df, site, clicks_zero_only=True)
    # "a": clicks=1>0 → KEPT unconditionally (regardless of thresholds)
    # "b": clicks=0, impressions=5<50 → EXCLUDED by threshold
    # "c": clicks=0, impressions=50>=50 ✓, position=3.0<=5 ✓ → KEPT
    assert sorted(list(filtered_zero_flag["query"])) == ["a", "c"]

    # Additional test: verify that clicks>0 rows are kept even if they fail thresholds
    df2 = pd.DataFrame(
        {
            "query": ["x", "y"],
            "impressions": [10, 100],  # "x" fails min_impressions
            "position": [2.0, 3.0],
            "clicks": [5, 0],  # "x" has clicks>0
        }
    )

    filtered_default = search.filter_by_thresholds(df2, site)
    # Default: only "y" passes thresholds
    assert list(filtered_default["query"]) == ["y"]

    filtered_clicks_zero = search.filter_by_thresholds(df2, site, clicks_zero_only=True)
    # clicks_zero_only: "x" kept despite failing threshold (because clicks>0), "y" kept (passes threshold)
    assert sorted(list(filtered_clicks_zero["query"])) == ["x", "y"]



"""
Functions for Google Analytics 3 (Universal Analytics) API
"""

import json
import logging
import re
import sys
from typing import Optional

import pandas as pd
from google.oauth2.credentials import Credentials
from googleapiclient import errors as err

from . import constants, errors, ga4, google_api, utils

LOGGER = logging.getLogger(__name__)


class MegatonUA(ga4.MegatonGA4):
    this = "Megaton UA"

    def __init__(self, credentials: Credentials, **kwargs):
        super().__init__(credentials, **kwargs)
        self.view = self.View(self)

    def _build_client(self):
        self.data_client = google_api.GoogleApi(
            "analyticsreporting",
            "v4",
            constants.DEFAULT_SCOPES,
            credentials=self.credentials,
        )
        self.admin_client = google_api.GoogleApi(
            "analytics",
            "v3",
            constants.DEFAULT_SCOPES,
            credentials=self.credentials,
        )

    def _update(self):
        """Returns account summaries accessible by the caller."""
        try:
            response = self.admin_client.management().accountSummaries().list().execute()
        except err.HttpError as e:
            status = e.resp.status
            data = json.loads(e.content.decode('utf-8'))
            message = data['error']['message']
            reason = data['error']['errors'][0]['reason']
            if status == 403:
                if 'not have any Google' in message or 'insufficientPermissions' in reason:
                    raise errors.NoDataReturned
                elif 'accessNotConfigured' in reason or 'disabled' in message:
                    raise errors.ApiDisabled(message, "Google Analytics API")
            raise e
        if response:
            results = []
            for i in response.get('items', []):
                if i['webProperties']:
                    account = {
                        'id': i['id'],
                        'name': i['name'],
                        'properties': [],
                    }
                    for p in i['webProperties']:
                        prop = {
                            'id': p['id'],
                            'name': p['name']
                        }
                        account['properties'].append(prop)
                    results.append(account)
            self.accounts = results
            return results

    class Account(ga4.MegatonGA4.Account):
        def __init__(self, parent):
            super().__init__(parent)

        def _update(self):
            response = self.parent.admin_client.management().webproperties().list(
                accountId=self.id).execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'id': i['id'],
                    'name': i['name'],
                    'industry': i.get('industryVertical', ''),
                    'service_level': i['level'],
                    'created_time': i['created'],
                    'updated_time': i['updated'],
                    'data_retention': i['dataRetentionTtl'],
                    'data_retention_reset_on_activity': i['dataRetentionResetOnNewActivity'],
                    'properties': '',
                }
                results.append(dict)
            self.properties = results
            return results

        def _clear(self):
            super()._clear()
            self.parent.view.clear()

        @property
        def segments(self):
            """Returns built-in and custom segments for the account."""
            response = self.parent.admin_client.management().segments().list().execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'id': i['id'],
                    'name': i['name'],
                    'type': i['type'],
                    'definition': i['definition'],
                }
                results.append(dict)
            return results

    class Property(ga4.MegatonGA4.Property):
        def __init__(self, parent):
            super().__init__(parent)

        def clear(self):
            super().clear()
            self.parent.view.clear()

        def _get_metadata(self):
            return {'dimensions': [], 'metrics': []}

        @property
        def custom_dimensions(self):
            """Returns custom dimensions for the property."""
            response = self.parent.admin_client.management().customDimensions().list(
                accountId=self.parent.account.id,
                webPropertyId=self.id
            ).execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'index': i['index'],
                    'display_name': i['name'],
                    'scope': i['scope'],
                    'active': i['active'],
                }
                results.append(dict)
            return results

        @property
        def custom_metrics(self):
            """Returns custom metrics for the property."""
            response = self.parent.admin_client.management().customMetrics().list(
                accountId=self.parent.account.id,
                webPropertyId=self.parent.property.id
            ).execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'index': i['index'],
                    'display_name': i['name'],
                    'scope': i['scope'],
                    'type': i['type'],
                    'active': i['active'],
                }
                results.append(dict)
            return results

        def _update(self):
            # self.clear()
            self.get_info()
            # self.get_available()  # Metadata API is not implemented

            # get views
            response = self.parent.admin_client.management().profiles().list(
                accountId=self.parent.account.id,
                webPropertyId=self.parent.property.id
            ).execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'property_id': i['webPropertyId'],
                    'id': i['id'],
                    'name': i['name'],
                    'currency': i['currency'],
                    'time_zone': i['timezone'],
                    'url': i['websiteUrl'],
                    'type': i['type'],
                    'ecommerce': i['eCommerceTracking'],
                    'bot_filtering': i.get('botFilteringEnabled', False),
                    'site_search_parameters': i.get('siteSearchQueryParameters', ''),
                    'default_page': i.get('defaultPage', ''),
                    'created_time': i['created'],
                    'updated_time': i['updated'],
                }
                results.append(dict)
            self.views = results
            # return results

        def get_info(self):
            """Get property data from parent account"""
            dict = [p for p in self.parent.account.properties if p['id'] == self.id][0]
            self.name = dict['name']
            self.created_time = dict['created_time']
            self.updated_time = dict['updated_time']
            self.industry = dict['industry']
            self.service_level = dict['service_level']
            self.data_retention = dict.get('data_retention', '')
            self.data_retention_reset_on_activity = dict.get('data_retention_reset_on_activity', '')
            return dict

        def get_dimensions(self):
            """Get custom dimension settings"""
            if not self.api_custom_dimensions:
                self.api_custom_dimensions = self.custom_dimensions()
            return self.api_custom_dimensions

        def get_metrics(self):
            """Get custom metrics settings"""
            if not self.api_custom_metrics:
                self.api_custom_metrics = self.custom_metrics()
            return self.api_custom_metrics

        def show(self, me: str = 'info', index_col: Optional[str] = None):
            res = None
            sort_values = []
            if me == 'custom_dimensions':
                index_col = 'index'
                res = self.get_dimensions()
            elif me == 'custom_metrics':
                index_col = 'index'
                res = self.get_metrics()

            if res:
                if index_col:
                    return pd.DataFrame(res).set_index(index_col).sort_values(by=sort_values)
                else:
                    return pd.DataFrame(res).sort_values(by=sort_values)
            return pd.DataFrame()

    class View:
        def __init__(self, parent):
            self.parent = parent
            self.id = None
            self.name = None
            self.currency = None
            self.time_zone = None
            self.url = None
            self.type = None
            self.ecommerce = None
            self.bot_filtering = None
            self.site_search_parameters = None
            self.default_page = None
            self.created_time = None
            self.updated_time = None

        def clear(self):
            self.id = None
            self.name = None
            self.currency = None
            self.time_zone = None
            self.url = None
            self.type = None
            self.ecommerce = None
            self.bot_filtering = None
            self.site_search_parameters = None
            self.default_page = None
            self.created_time = None
            self.updated_time = None

        def select(self, id: str):
            if id:
                if id != self.id:
                    self.id = id
                    state = getattr(self.parent, "_state", None)
                    if state is not None:
                        state.ga_version = getattr(self.parent, "_ga_version", None)
                        state.ga_view_id = id
                    self._update()
            else:
                self.clear()
                state = getattr(self.parent, "_state", None)
                if state is not None:
                    state.ga_version = getattr(self.parent, "_ga_version", None)
                    state.ga_view_id = None

        def _update(self):
            self.get_info()

        def get_info(self):
            """Get view data from parent property"""
            i = [p for p in self.parent.property.views if p['id'] == self.id][0]
            self.name = i['name']
            self.currency = i['currency']
            self.time_zone = i['time_zone']
            self.url = i['url']
            self.type = i['type']
            self.ecommerce = i['ecommerce']
            self.bot_filtering = i['bot_filtering']
            self.site_search_parameters = i['site_search_parameters']
            self.default_page = i['default_page']
            self.created_time = i['created_time']
            self.updated_time = i['updated_time']
            return i

        @property
        def goals(self):
            """Get Goals"""
            response = self.parent.admin_client.management().goals().list(
                accountId=self.parent.account.id,
                webPropertyId=self.parent.property.id,
                profileId=self.id
            ).execute()
            results = []
            for i in response.get('items', []):
                dict = {
                    'id': i.get('id'),
                    'name': i.get('name'),
                    'value': i.get('value'),
                    'type': i.get('type'),
                    'url_destination': i.get('urlDestinationDetails'),
                    'time_on_site': i.get('visitTimeOnSiteDetails'),
                    'pages_per_session': i.get('visitNumPagesDetails'),
                    'event': i.get('eventDetails'),
                    'created_time': i.get('created'),
                    'updated_time': i.get('updated'),
                    'active': i.get('active'),
                }
                results.append(dict)
            return results

        def show(self, me: str = 'info', index_col: Optional[str] = None):
            res = None
            sort_values = []
            if me == 'info':
                res = [self.get_info()]
                index_col = 'id'
            elif me == 'goals':
                res = self.goals
                index_col = 'id'

            if res:
                if index_col:
                    return pd.DataFrame(res).set_index(index_col).sort_values(by=sort_values)
                else:
                    return pd.DataFrame(res).sort_values(by=sort_values)
            return pd.DataFrame()

    class Report(ga4.MegatonGA4.Report):

        def _format_name(self, name: str, type: str = None):
            """Convert api_name to an object for request and OrderBy"""
            after = ("ga:" + name) if not name.startswith("ga:") else name
            if type == 'dimension':
                # Dimension object in ReportRequest
                return {
                    'name': after,
                }
            elif type == 'metric':
                # Metric object in ReportRequest
                return {
                    'expression': after,
                    'alias': name,
                }
            else:
                # simple api_name for OrderBy in ReportRequest, etc
                return after

        def _parse_operator(self, operator: str, type: str):
            """Convert a string to legacy filter operator from Core Reporting API v3"""
            is_dimension = True if type.startswith('d') else False

            if is_dimension:
                if operator.endswith('='):
                    return 'EXACT'
                elif operator.endswith('~'):
                    return 'REGEXP'
                return 'PARTIAL'
            else:  # is metric
                if operator.endswith('='):
                    return 'EQUAL'
                elif operator == '>':
                    return 'GREATER_THAN'
                elif operator == '<':
                    return 'LESS_THAN'

        def _parse_filter_condition(self, condition: str, **kwargs):
            """Split a string into name, operator and value
            Operator:
                == : Equals
                != : Does not equal
                =@ : Contains substring
                !@ : Does not contain substring
                =~ : Contains a match for the regular expression
                !~ : Does not match regular expression
                > : Greater than
                < : Less than
            """
            m = re.search(r'^(.+)(==|!=|=@|!@|=~|!~|>|<)(.+)$', condition)
            if m:
                field = self._format_name(m.groups()[0])
                type = kwargs.get('type')
                value = m.groups()[2]
                op = m.groups()[1]
                is_not = True if op.startswith('!') else False
                operator = self._parse_operator(op, type)

                if type == 'dimensions':
                    return {
                        'dimensionName': field,
                        'not': is_not,
                        'operator': operator,
                        'expressions': [value],
                        'caseSensitive': False
                    }
                elif type == 'metrics':
                    return {
                        'metricName': field,
                        'not': is_not,
                        'operator': operator,
                        'comparisonValue': value
                    }

        def _format_filter(self, conditions: str, type='dimensions'):
            """Convert legacy filters format from Core Reporting API v3
             to DimensionFilterClause or MetricFilterClause object"""
            if not conditions:
                return []

            filters = [self._parse_filter_condition(i, type=type) for i in conditions.split(';')]
            return [{
                'operator': 'AND',
                'filters': filters
            }]

        def _format_order_bys(self, before: str):
            """Convert legacy sort format from Core Reporting API v3 to a list of OrderBy object"""
            if not before:
                return

            result = []
            for i in before.split(','):
                obj = {}
                try:
                    _, field = i.split('-')
                except ValueError:
                    # Ascending
                    obj['fieldName'] = self._format_name(i)
                    obj['sortOrder'] = 'ASCENDING'
                else:
                    # Descending
                    obj['fieldName'] = self._format_name(field)
                    obj['sortOrder'] = 'DESCENDING'
                result.append(obj)
            return result

        def _format_request(self, **kwargs):
            """Construct a request for API"""
            dimension_api_names = [self._format_name(r, type='dimension') for r in kwargs.get('dimensions')]
            metrics_api_names = [self._format_name(r, type='metric') for r in kwargs.get('metrics')]

            return {
                'viewId': self.parent.view.id,
                'dateRanges': [{
                    'startDate': kwargs.get('start_date'),
                    'endDate': kwargs.get('end_date')
                }],
                'samplingLevel': 'LARGE',
                'dimensions': dimension_api_names,
                'dimensionFilterClauses': self._format_filter(kwargs.get('dimension_filter'), type='dimensions'),
                'metrics': metrics_api_names,
                'metricFilterClauses': self._format_filter(kwargs.get('metric_filter'), type='metrics'),
                'orderBys': self._format_order_bys(kwargs.get('order_bys')),
                'segments': kwargs.get('segments'),
                'includeEmptyRows': False,
                'hideTotals': not kwargs.get('show_total', False),
                'hideValueRanges': True,
                'pageSize': kwargs.get('limit'),
            }

        def _parse_response(self, report: dict):
            all_data = []
            names = []
            dimension_types = []
            metric_types = []

            for i in report['columnHeader']['dimensions']:
                names.append(i.replace('ga:', ''))
                dimension_types.append('category')

            for i in report['columnHeader']['metricHeader']['metricHeaderEntries']:
                names.append(i['name'])
                metric_types.append(i['type'])

            for row in report['data'].get('rows', []):
                row_data = []
                for d in row['dimensions']:
                    row_data.append(d)
                for d in row['metrics']:
                    for i, v in enumerate(d['values']):
                        row_data.append(self._convert_metric(v, metric_types[i]))
                all_data.append(row_data)

            return all_data, names, dimension_types + metric_types

        def _request_report_api(self, offset: str, request: dict):
            if offset:
                request["pageToken"] = offset

            try:
                response = self.parent.data_client.reports().batchGet(
                    body={
                        'reportRequests': [request],
                        'useResourceQuotas': False  # only for 360
                    }
                ).execute()
            except err.HttpError as e:
                code = e.resp.get('code')
                message = ''
                reason = ''
                status = ''
                try:
                    data = json.loads(e.content.decode('utf-8'))
                    code = data['error']["code"]
                    message = data['error']['message']
                    reason = data['error']['errors'][0]['reason']
                    status = data['error']['status']
                except:  # noqa
                    pass
                if code == 400:
                    if status == 'INVALID_ARGUMENT' or 'nknown ' in message:
                        raise errors.BadRequest(message)
                elif status == 403:
                    if 'not have any Google' in message or 'insufficientPermissions' in reason:
                        raise errors.NoDataReturned
                    elif 'accessNotConfigured' in reason or 'disabled' in message:
                        raise errors.ApiDisabled(message, "Google Analytics API")
                raise e

            report_data = response.get('reports', [])[0]
            total_rows = report_data['data'].get('rowCount', 0)

            samples_count = report_data['data'].get('samplesReadCounts')
            samples_size = report_data['data'].get('samplingSpaceSizes')
            next_token = report_data.get('nextPageToken', None)
            if samples_count:
                LOGGER.warn(f"samplesReadCounts = {samples_count}")
            if samples_size:
                LOGGER.warn(f"samplingSpaceSizes = {samples_size}")

            data, headers, types = self._parse_response(report_data)

            return data, total_rows, headers, types, next_token

        def _report_generator(self, request: dict):
            """Send request to get report data"""
            if not self.parent.view.id:
                # LOGGER.error("Viewを先に選択してから実行してください。")
                return

            token = "0"
            while True:
                try:
                    (data, total_rows, headers, types, next_token) = self._request_report_api(token, request)
                except err.HttpError as e:
                    value = str(sys.exc_info()[1])
                    if 'disabled' in value:
                        LOGGER.error("\nGCPのプロジェクトでAnalytics Reporting APIを有効化してください。")
                        raise errors.ApiDisabled
                    elif 'Invalid value' in value:
                        LOGGER.error("レポート抽出条件の書式や内容を見直してください。")
                        raise errors.BadRequest
                    else:
                        raise e

                if token == "0" and total_rows:
                    LOGGER.info(f"Total {total_rows} rows found.")

                    if total_rows == 10001:
                        if [d for d in request['dimensions'] if d['name'] == 'ga:clientId']:
                            LOGGER.info("clientId is not officially supported by Google. Using this dimension in an "
                                        "Analytics report may thus result in unexpected & unexplainable behavior ("
                                        "such as restricting the report to exactly 10,000 or 10,001 rows).")
                    self.headers = headers
                    self.types = types

                if len(data):
                    LOGGER.info(f"(received row #{token}-{int(token) + len(data) - 1})")

                for row in data:
                    yield row

                if not next_token:
                    # no more data
                    break
                if len(data) == 0:
                    # API bug
                    raise errors.PartialDataReturned()

                token = next_token

        def run(self, dimensions: list, metrics: list, return_generator: Optional[bool] = None, **kwargs):
            """Get Analytics report data"""
            if not self.parent.view.id:
                LOGGER.error("Viewを先に選択してから実行してください。")
                return

            if len(dimensions) > 7:
                LOGGER.warn("Up to 7 dimensions are allowed.")
                dimensions = dimensions[:7]
            if len(metrics) > 10:
                LOGGER.warn("Up to 10 dimensions are allowed.")
                metrics = metrics[:10]

            start_date = kwargs.get('start_date', self.start_date)
            end_date = kwargs.get('end_date', self.end_date)
            LOGGER.info(f"Requesting a report ({start_date} - {end_date})")

            request = self._format_request(
                dimensions=dimensions,
                metrics=metrics,
                start_date=start_date,
                end_date=end_date,
                dimension_filter=kwargs.get('dimension_filter'),
                metric_filter=kwargs.get('metric_filter'),
                order_bys=kwargs.get('order_bys'),
                segments=kwargs.get('segments'),
                show_total=False,
                limit=kwargs.get('limit', 10000),
            )
            # print(request)

            iterator = self._report_generator(request)

            if return_generator:
                return iterator

            try:
                return pd.DataFrame(list(iterator), columns=self.headers)
            except AttributeError:
                LOGGER.info("No data found.")
                return pd.DataFrame()
            except errors.PartialDataReturned:
                LOGGER.warn("APIのバグにより全データを取得できませんでした。")
                if start_date != end_date:
                    LOGGER.warn("1日毎にデータを分割して取得します。")
                    all_data = []
                    for date in utils.get_date_range(start_date, end_date):
                        print(f"...{date}", end='')
                        request['dateRanges'][0]['startDate'] = date
                        request['dateRanges'][0]['endDate'] = date
                        request['pageToken'] = "0"
                        iterator = self._report_generator(request)
                        data = list(iterator)
                        LOGGER.debug(f"{len(data)} rows")
                        all_data.extend(data)
                    return pd.DataFrame(all_data, columns=self.headers)


def get_cid_date_page(ga3, conf: dict):
    """データ抽出
        Dimensions:
            対象ページを閲覧したclient_id,セッションカウント, 日付
        Metrics:
            入口, PV, 出口
    """
    filter = []
    if conf['include_domains']:
        filter.append(f"hostname=~{conf['include_domains']}")
    if conf['include_pages']:
        filter.append(f"pagePath=~{conf['include_pages']}")
    if conf['exclude_pages']:
        filter.append(f"pagePath!~{conf['exclude_pages']}")
    dimension_filter = ";".join(filter)

    df = ga3.report.show(
        dimensions=[
            'clientId',
            'sessionCount',
            'date',
            'pagePath',
        ],
        dimension_filter=dimension_filter,
        metrics=[
            'entrances',
            'uniquePageviews',
            'exits',
        ],
        order_bys='clientId,sessionCount,date,pagePath',
    )

    if len(df):
        # 値を変換
        if conf['page_regex']:
            utils.replace_columns(df, [('pagePath', conf['page_regex'], '')])
        # カラム名を変更
        return df.rename(columns={
            'pagePath': 'page',
            'uniquePageviews': 'sessions',
        })
    else:
        raise errors.NoDataReturned


def get_last_returned_date(ga3):
    """データ抽出：再訪問した人の最終訪問日
        Dimensions:
            再訪問時のclient_id,日付
        Metrics:
            入口
    """
    df = ga3.report.show(
        dimensions=[
            'clientId',
            'date',
        ],
        dimension_filter='ga:sessionCount!=1',
        metrics=[
            'entrances',  # 不要
        ],
        order_bys='ga:clientId,ga:date',
    )

    # cid単位でまとめて最後に訪問した日を算出
    _df = df.groupby(['clientId']).max().rename(columns={'date': 'last_visit_date'})

    return _df.drop(['entrances'], inplace=False, axis=1)


def get_no_entrance_cv_cid(ga3, include_pages=None, metric_filter: str = 'ga:entrances<1'):
    """元データ抽出：入口以外でCVページに到達したcid
    """
    filter = []
    if include_pages:
        filter.append(f'pagePath=~{include_pages}')
    dimension_filter = ";".join(filter)

    _df = ga3.report.show(
        dimensions=[
            'pagePath',  # 不要（確認のため）
            'clientId',
            'date',
            'sessionCount',
        ],
        dimension_filter=dimension_filter,
        metrics=[
            'users',  # 不要
        ],
        metric_filter=metric_filter,
        order_bys='pagePath,clientId,date',
    )
    if len(_df):
        return _df.drop(['users'], axis=1)
    else:
        return pd.DataFrame()


def get_page_title(ga3, conf: dict):
    """ 対象・除外・変換条件を指定してpageとtitleを抽出 """
    filter = []
    if conf['include_domains']:
        filter.append(f"hostname=~{conf['include_domains']}")
    if conf['include_pages']:
        filter.append(f"pagePath=~{conf['include_pages']}")
    if conf['exclude_pages']:
        filter.append(f"pagePath!~{conf['exclude_pages']}")
    dimension_filter = ";".join(filter)

    df = ga3.report.show(
        dimensions=[
            'pagePath',
            'pageTitle',
        ],
        dimension_filter=dimension_filter,
        metrics=[
            'uniquePageviews',
        ],
        order='-uniquePageviews',
    )

    df.rename(columns={
        'pagePath': 'page',
        'pageTitle': 'title',
    }, inplace=True)

    # 値を変換
    if conf['page_regex']:
        utils.replace_columns(df, [('pagePath', conf['page_regex'], '')])
    if conf['title_regex']:
        utils.replace_columns(df, [('title', conf['title_regex'], '')])

    # group byでまとめる
    try:
        return df.sort_values(
            ['page', 'uniquePageviews'],
            ascending=False
        ).groupby('page').first().reset_index()[['page', 'title']]
    except KeyError:
        raise errors.NoDataReturned



import pytest
import requests

from megaton import errors
from megaton.gsheet import MegatonGS


def _new_gs():
    gs = MegatonGS.__new__(MegatonGS)
    return gs


def test_resolve_timeout_default(monkeypatch):
    monkeypatch.delenv("MEGATON_GS_TIMEOUT", raising=False)
    gs = _new_gs()
    assert gs._resolve_timeout(None) == 180.0


def test_resolve_timeout_env_override(monkeypatch):
    monkeypatch.setenv("MEGATON_GS_TIMEOUT", "45")
    gs = _new_gs()
    assert gs._resolve_timeout(None) == 45.0


def test_resolve_timeout_env_disable(monkeypatch):
    monkeypatch.setenv("MEGATON_GS_TIMEOUT", "0")
    gs = _new_gs()
    assert gs._resolve_timeout(None) is None


def test_open_raises_timeout():
    gs = _new_gs()

    class DummyClient:
        def open_by_url(self, url):
            raise requests.exceptions.Timeout("boom")

    gs._client = DummyClient()

    with pytest.raises(errors.Timeout):
        gs.open("https://docs.google.com/spreadsheets/d/xxxx")


def test_open_raises_request_error():
    gs = _new_gs()

    class DummyClient:
        def open_by_url(self, url):
            raise requests.exceptions.RequestException("boom")

    gs._client = DummyClient()

    with pytest.raises(errors.RequestError):
        gs.open("https://docs.google.com/spreadsheets/d/xxxx")



from __future__ import annotations

import re
from urllib.parse import urlparse


def classify_by_regex(df, src_col, mapping, out_col, default="other"):
    if src_col not in df.columns:
        raise ValueError(f"Missing source column: {src_col}")

    def _classify(value):
        if not isinstance(value, str):
            return default
        for pattern, label in mapping.items():
            try:
                if re.search(pattern, value):
                    return label
            except re.error:
                continue
        return default

    result = df.copy()
    result[out_col] = result[src_col].apply(_classify)
    return result


def infer_label_by_domain(series, domain_to_label_map, default="不明"):
    if series is None:
        return series

    def _infer(value):
        if not isinstance(value, str) or not value:
            return default
        domain = urlparse(value).netloc if "://" in value else value
        domain = domain.lower().strip()
        if domain.startswith("www."):
            domain = domain[4:]
        for key, label in domain_to_label_map.items():
            if key.lower() in domain:
                return label
        return default

    return series.apply(_infer)



"""Tests for clean=True parameter in mg.search.run()"""
import pytest
import pandas as pd
from megaton.start import SearchResult


def test_clean_aggregation_with_weighted_position():
    """SearchResult の clean 相当の処理が position を重み付き平均で集計する"""
    
    # URL 正規化前のデータ
    df = pd.DataFrame({
        'page': [
            'https://example.com/Page?utm=123',
            'https://EXAMPLE.COM/page?utm=456',  # 異なる大文字小文字とパラメータ
        ],
        'clicks': [10, 20],
        'impressions': [100, 200],
        'position': [3.0, 6.0]
    })
    
    result = SearchResult(df, None, ['page'])
    
    # clean 相当の処理: decode → remove_params → remove_fragment → lower
    cleaned = (result
               .decode(group=False)
               .remove_params(group=False)
               .remove_fragment(group=False)
               .lower(columns=['page'], group=True))  # 最後に group=True で集計
    
    # page が正規化されて1行に集計される
    assert len(cleaned.df) == 1
    
    # clicks と impressions は合計
    assert cleaned.df['clicks'].iloc[0] == 30
    assert cleaned.df['impressions'].iloc[0] == 300
    
    # position は重み付き平均: (3.0 * 100 + 6.0 * 200) / 300 = 5.0
    assert cleaned.df['position'].iloc[0] == 5.0
    
    # page が正規化されている（小文字、パラメータなし）
    assert 'example.com/page' in cleaned.df['page'].iloc[0]


def test_clean_without_position():
    """position がない場合も正しく動作"""
    
    df = pd.DataFrame({
        'page': [
            'https://example.com/Page',
            'https://EXAMPLE.COM/page',
        ],
        'clicks': [10, 20],
        'impressions': [100, 200]
        # position なし
    })
    
    result = SearchResult(df, None, ['page'])
    cleaned = result.lower(columns=['page'], group=True)
    
    # 1行に集計される
    assert len(cleaned.df) == 1
    assert cleaned.df['clicks'].iloc[0] == 30
    assert cleaned.df['impressions'].iloc[0] == 300
    
    # position 列が存在しない
    assert 'position' not in cleaned.df.columns


def test_clean_with_ctr():
    """CTR が集計後に正しく再計算される"""
    
    df = pd.DataFrame({
        'page': [
            'https://example.com/Page',
            'https://EXAMPLE.COM/page',
        ],
        'clicks': [10, 20],
        'impressions': [100, 300],
        'ctr': [0.1, 0.0666667]  # GSC が返す元の CTR
    })
    
    result = SearchResult(df, None, ['page'])
    cleaned = result.lower(columns=['page'], group=True)
    
    # 1行に集計される
    assert len(cleaned.df) == 1
    assert cleaned.df['clicks'].iloc[0] == 30
    assert cleaned.df['impressions'].iloc[0] == 400
    
    # CTR が再計算されている: 30 / 400 = 0.075
    assert 'ctr' in cleaned.df.columns
    expected_ctr = 30 / 400
    assert abs(cleaned.df['ctr'].iloc[0] - expected_ctr) < 0.001



from dataclasses import dataclass
from typing import Optional


@dataclass
class MegatonState:
    """Lightweight mirror of current selections/settings."""
    ga_version: Optional[str] = None
    ga_account_id: Optional[str] = None
    ga_property_id: Optional[str] = None
    ga_view_id: Optional[str] = None
    gs_url: Optional[str] = None
    gs_title: Optional[str] = None
    gs_sheet_name: Optional[str] = None
    bq_project_id: Optional[str] = None
    bq_dataset_id: Optional[str] = None
    bq_table_id: Optional[str] = None
    headless: Optional[bool] = None



from datetime import datetime
from types import SimpleNamespace

import pandas as pd

from megaton import ga4


def _parent_with_property(property_id="123"):
    return SimpleNamespace(
        property=SimpleNamespace(
            id=property_id,
            created_time=datetime(2024, 1, 1),
            api_metadata={
                "dimensions": [
                    {"api_name": "date", "display_name": "Date"},
                    {"api_name": "country", "display_name": "Country"},
                ],
                "metrics": [
                    {"api_name": "eventCount", "display_name": "Event count"},
                    {"api_name": "totalUsers", "display_name": "Total users"},
                ],
            },
        ),
        data_client=SimpleNamespace(),
    )


def test_parse_operator_metric_variants():
    report = ga4.MegatonGA4.Report(_parent_with_property())

    assert report._parse_operator("==", "metrics").name == "EQUAL"
    assert report._parse_operator(">", "metrics").name == "GREATER_THAN"
    assert report._parse_operator("<", "metrics").name == "LESS_THAN"
    assert report._parse_operator("~", "metrics").name == "OPERATION_UNSPECIFIED"


def test_format_filter_multiple_conditions_builds_and_group():
    report = ga4.MegatonGA4.Report(_parent_with_property())

    expr = report._format_filter("date==2024-01-01;eventCount>10")

    assert expr.and_group is not None
    assert len(expr.and_group.expressions) == 2
    assert expr.and_group.expressions[0].filter.field_name == "date"
    assert expr.and_group.expressions[1].filter.field_name == "eventCount"


def test_convert_metric_numeric_and_fallback():
    report = ga4.MegatonGA4.Report(_parent_with_property())

    assert report._convert_metric("10", "TYPE_INTEGER") == 10
    assert report._convert_metric("1.5", "TYPE_FLOAT") == 1.5
    assert report._convert_metric("not-number", "TYPE_INTEGER") == "not-number"


def test_parse_response_converts_metric_values():
    report = ga4.MegatonGA4.Report(_parent_with_property())

    response = SimpleNamespace(
        dimension_headers=[SimpleNamespace(name="date")],
        metric_headers=[SimpleNamespace(name="eventCount", type_=1)],
        rows=[
            SimpleNamespace(
                dimension_values=[SimpleNamespace(value="2024-01-01")],
                metric_values=[SimpleNamespace(value="7")],
            )
        ],
    )

    data, headers, types = report._parse_response(response)

    assert data == [["2024-01-01", 7]]
    assert headers == ["date", "eventCount"]
    assert types == ["category", "TYPE_INTEGER"]


def test_run_returns_none_without_property_id():
    report = ga4.MegatonGA4.Report(_parent_with_property(property_id=None))

    assert report.run(dimensions=["date"], metrics=["eventCount"]) is None


def test_run_paginates_and_returns_dataframe(monkeypatch):
    report = ga4.MegatonGA4.Report(_parent_with_property())
    offsets = []

    monkeypatch.setattr(report, "_format_request", lambda **_kwargs: SimpleNamespace())

    def _fake_request(offset, _request, **_kwargs):
        offsets.append(offset)
        if offset == 0:
            return (
                [["2024-01-01", 1], ["2024-01-02", 2]],
                3,
                ["date", "eventCount"],
                ["category", "TYPE_INTEGER"],
            )
        if offset == 2:
            return (
                [["2024-01-03", 3]],
                3,
                ["date", "eventCount"],
                ["category", "TYPE_INTEGER"],
            )
        return (
            [],
            3,
            ["date", "eventCount"],
            ["category", "TYPE_INTEGER"],
        )

    monkeypatch.setattr(report, "_request_report_api", _fake_request)
    df = report.run(dimensions=["date"], metrics=["eventCount"], limit=2)

    assert offsets == [0, 2]
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 3
    assert list(df.columns) == ["date", "eventCount"]


def test_run_trims_dimensions_and_metrics_before_request(monkeypatch):
    report = ga4.MegatonGA4.Report(_parent_with_property())
    captured = {}

    def _capture_request(**kwargs):
        captured["dimensions"] = kwargs["dimensions"]
        captured["metrics"] = kwargs["metrics"]
        return SimpleNamespace()

    monkeypatch.setattr(report, "_format_request", _capture_request)
    monkeypatch.setattr(report, "_request_report_api", lambda *_args, **_kwargs: ([], 0, [], []))

    dims = [f"d{i}" for i in range(11)]
    metrics = [f"m{i}" for i in range(12)]
    df = report.run(dimensions=dims, metrics=metrics, limit=10)

    assert isinstance(df, pd.DataFrame)
    assert df.empty
    assert len(captured["dimensions"]) == 9
    assert len(captured["metrics"]) == 10


def test_convert_ga4_type_to_bq_type_mapping():
    assert ga4.convert_ga4_type_to_bq_type("string") == "STRING"
    assert ga4.convert_ga4_type_to_bq_type("int") == "INT64"
    assert ga4.convert_ga4_type_to_bq_type("integer") == "INT64"
    assert ga4.convert_ga4_type_to_bq_type("float") == "FLOAT"
    assert ga4.convert_ga4_type_to_bq_type("double") == "FLOAT"
    assert ga4.convert_ga4_type_to_bq_type("other") is None


def test_convert_proto_datetime_supports_seconds_and_timestamp():
    class _WithSeconds:
        seconds = 0

    class _WithTimestamp:
        def timestamp(self):
            return 0

    dt1 = ga4.convert_proto_datetime(_WithSeconds())
    dt2 = ga4.convert_proto_datetime(_WithTimestamp())

    assert dt1.year == 1970
    assert dt2.year == 1970



import pandas as pd
from unittest.mock import MagicMock

from megaton.start import Megaton


def test_save_to_sheet_passes_options():
    app = Megaton()
    app._sheets = MagicMock()
    df = pd.DataFrame({"a": [2, 1]})

    app.save.to.sheet(
        "Sheet1",
        df,
        sort_by="a",
        sort_desc=False,
        start_row=2,
        create_if_missing=True,
        auto_width=True,
        freeze_header=True,
    )

    app._sheets.save_sheet.assert_called_once_with(
        "Sheet1",
        df,
        sort_by="a",
        sort_desc=False,
        start_row=2,
        create_if_missing=True,
        auto_width=True,
        freeze_header=True,
    )


def test_sheet_save_passes_options():
    app = Megaton()
    app._sheets = MagicMock()
    app.sheet._ensure_spreadsheet = MagicMock()
    app.sheet._ensure_sheet_selected = MagicMock(return_value="Sheet2")
    df = pd.DataFrame({"a": [1]})

    app.sheet.save(
        df,
        sort_by="a",
        start_row=3,
        auto_width=True,
        freeze_header=True,
    )

    app._sheets.save_sheet.assert_called_once_with(
        "Sheet2",
        df,
        sort_by="a",
        sort_desc=True,
        start_row=3,
        auto_width=True,
        freeze_header=True,
    )


def test_append_to_sheet_passes_options():
    app = Megaton()
    app._sheets = MagicMock()
    df = pd.DataFrame({"a": [1]})

    app.append.to.sheet(
        "Sheet1",
        df,
        create_if_missing=True,
        auto_width=True,
        freeze_header=True,
    )

    app._sheets.append_sheet.assert_called_once_with(
        "Sheet1",
        df,
        create_if_missing=True,
        auto_width=True,
        freeze_header=True,
    )


def test_upsert_to_sheet_passes_options():
    app = Megaton()
    app._sheets = MagicMock()
    app.state.gs_url = "https://example.com/sheet"
    df = pd.DataFrame({"a": [1]})

    app.upsert.to.sheet(
        "Sheet1",
        df,
        keys=["a"],
        auto_width=True,
        freeze_header=True,
    )

    app._sheets.upsert_df.assert_called_once_with(
        "https://example.com/sheet",
        "Sheet1",
        df,
        keys=["a"],
        columns=None,
        sort_by=None,
        create_if_missing=True,
        auto_width=True,
        freeze_header=True,
    )



"""
Functions for Google API
"""

import json
import logging
import time

from googleapiclient import errors
from googleapiclient.discovery import build, DISCOVERY_URI

_REQUIRED_CONFIG_KEYS = frozenset(("auth_uri", "token_uri", "client_id"))


class GoogleApi(object):
    """Google API helper object"""

    def __init__(self, api="oauth2", version="v2", scopes=None, **kwargs):
        """constructor"""
        if scopes is None:
            scopes = ['https://www.googleapis.com/auth/analytics.readonly']
        self.api = api
        self.api_version = version
        self.scopes = scopes
        self.credentials = kwargs.get('credentials')
        self._service = None
        self.discovery_url = kwargs.get('discovery_url', DISCOVERY_URI)
        self.retries = kwargs.get('retries', 3)
        self.log = logging.getLogger("__name__")

    @property
    def service(self):
        """get or create a api service"""
        if self._service is None:
            # self.log.debug(f"Creating a service for {self.api} API")
            self._service = build(self.api,
                                  self.api_version,
                                  credentials=self.credentials,
                                  cache_discovery=False,
                                  discoveryServiceUrl=self.discovery_url)
        return self._service

    def retry(self, service_method, retry_count=0):
        """retry a google api call and check for rate limits
        """
        try:
            return service_method.execute(num_retries=retry_count)
        except errors.HttpError as e:
            code = e.resp.get('code')
            # reason = ''
            message = ''
            data = None
            try:
                data = json.loads(e.content.decode('utf-8'))
                code = data['error']["code"]
                message = data['error']['message']
                # reason = data['error']['errors'][0]['reason']
            except:  # noqa
                data = e.content
            if code == 403 and "rate limit exceeded" in message.lower():
                self.log.debug("rate limit reached, sleeping for %s seconds", 2 ** retry_count)
                time.sleep(2 ** retry_count)
                return self.retry(service_method, retry_count + 1)
            else:
                self.log.debug(f"got HttpError (content={data}")
                raise
        except BrokenPipeError:
            self.log.debug("BrokenPipeError occurred but attempting to retry")
            return self.retry(service_method, retry_count + 1)
        except KeyboardInterrupt:
            raise
        except:  # noqa
            self.log.exception("Failed to execute api method")
            raise

    def __getattr__(self, name):
        """get attribute or service wrapper
        :param name: attribute / service name
        :return:
        """
        return getattr(MethodHelper(self, self.service), name)


class MethodHelper(object):
    """helper to streamline api calls"""

    def __init__(self, google_api, service, name=None, path=None):
        """create a method helper
        :param google_api GoogleApi instance of api
        :param service Google API service (GoogleApi.service) or method of it
        :param name method name
        :param path API path i.e. for compute: instances.list
        """
        self.google_api = google_api
        self.service = service
        self.name = name
        self.path = path if path is not None else []
        if name is not None:
            self.path.append(name)
        # print("constructor %s", name)

    def execute(self, *args, **kwargs):
        """execute service api"""
        # self.log.info("execute %s", self.name)
        return self.google_api.retry(self.service)

    def call(self, *args, **kwargs):
        """wrapper for service methods
        this wraps an GoogleApi.service call so the next level can also use helpers
        i.e. for compute v1 api GoogleApi.service.instances() can be used as Google.instances()
        and will return a MethodHelper instance
        """
        # self.log.info("call %s", self.name)
        return MethodHelper(self.google_api, getattr(self.service, self.name)(*args, **kwargs))

    def __getattr__(self, name):
        """get service method"""
        # self.log.info("getattr %s", name)
        if not hasattr(self.service, name):
            err_msg = u"API method {} unknown on {} {}".format(u".".join(self.path + [name]),
                                                               self.google_api.api,
                                                               self.google_api.api_version)
            raise RuntimeError(err_msg)
        return MethodHelper(self.google_api, self.service, name, self.path).call



"""
Functions for Google Cloud BigQuery
"""

from typing import List
import re
import sys

from google.api_core.exceptions import PermissionDenied
from google.cloud import bigquery
from google.cloud import bigquery_datatransfer
from google.cloud.exceptions import NotFound

from . import constants, ga4


class MegatonBQ:
    """Class for Google Cloud BigQuery client
    """

    def __init__(self, parent, credentials, project_id: str):
        self.parent = parent
        self.id = project_id
        self.datasets = None
        self.credentials = credentials
        self.client = bigquery.Client(
            project=self.id,
            credentials=self.credentials
        )
        self.dts_client = bigquery_datatransfer.DataTransferServiceClient(
            credentials=self.credentials
        )
        self.dataset = self.Dataset(self)
        self.table = self.Table(self)
        self.ga4 = self.GA4(self)
        self.update()

    def update(self) -> None:
        """Get a list of dataset ids for the project"""
        # Make an API request.
        datasets = list(self.client.list_datasets())

        if datasets:
            # extract dataset id
            self.datasets = [d.dataset_id for d in datasets]
        else:
            print(f"project {self.id} has no datasets.")

    def run(self, query: str, to_dataframe: bool = False):
        """Run a query and return data
        Args:
            query (str):
                SQL query to be executed.
            to_dataframe (bool):
                if true, data is retured as pandas DataFrame
        """
        job = self.client.query(query=query)
        results = job.result()  # Waits for job to complete.

        if to_dataframe:
            return results.to_dataframe()
        else:
            return results

    class Dataset:
        def __init__(self, parent):
            self.parent = parent
            self.ref = None
            self.instance = None
            self.id = None
            self.tables = None

        def _clear(self) -> None:
            self.ref = None
            self.instance = None
            self.id = None
            self.tables = None
            self.parent.table.select()
            if hasattr(self.parent.parent, "state"):
                self.parent.parent.state.bq_dataset_id = None
                self.parent.parent.state.bq_table_id = None

        def select(self, dataset_id: str) -> None:
            """select dataset"""
            if dataset_id:
                if dataset_id in self.parent.datasets:
                    if dataset_id != self.id:
                        self.update(dataset_id)
                else:
                    print(f"dataset {dataset_id} is not found in the project {self.parent.id}")
            else:
                self._clear()

        def update(self, dataset_id: str = ''):
            """Get a list of table ids for the dataset specified"""
            id = dataset_id if dataset_id else self.id

            try:
                dataset = self.parent.client.get_dataset(id)
                self.instance = dataset
                self.ref = dataset.reference
                self.id = id
                if hasattr(self.parent.parent, "state"):
                    self.parent.parent.state.bq_dataset_id = id
            except NotFound as e:
                if 'Not found: Dataset' in str(e):
                    print(f"Dataset {dataset_id} is not found in the project {self.parent.id}")
                return False

            # Make an API request
            tables = list(self.parent.client.list_tables(dataset))

            if tables:
                # extract table id
                self.tables = [d.table_id for d in tables]
            else:
                print(f"dataset {self.id} has no tables.")

    class Table:
        def __init__(self, parent):
            self.parent = parent
            self.ref = None
            self.instance = None
            self.id = None

        # def _get_info(self):
        #     """Get metadata for the table"""

        def select(self, table_id: str) -> None:
            """select a table"""
            if table_id:
                if table_id in self.parent.dataset.tables:
                    if table_id != self.id:
                        self.update(table_id)
                else:
                    print(f"table {table_id} is not found in the dataset {self.parent.dataset.id}")
            else:
                self.ref = None
                self.instance = None
                self.id = None
                if hasattr(self.parent.parent, "state"):
                    self.parent.parent.state.bq_table_id = None

        def update(self, table_id: str = ''):
            """Get an api reference for the table"""
            id = table_id if table_id else self.id
            if self.parent.dataset.ref:
                try:
                    table_ref = self.parent.dataset.ref.table(id)
                    self.ref = table_ref
                    self.instance = self.parent.client.get_table(self.ref)
                    self.id = id
                    if hasattr(self.parent.parent, "state"):
                        self.parent.parent.state.bq_table_id = id
                    # self._get_info()
                except Exception as e:
                    raise e
            else:
                print("Please select a dataset first.")

        def create(self, table_id: str, schema: bigquery.SchemaField, description: str = '',
                   partitioning_field: str = '', clustering_fields=None) -> bigquery.table.Table:
            if clustering_fields is None:
                clustering_fields = []
            dataset_ref = self.parent.dataset.ref
            table_ref = dataset_ref.table(table_id)
            table = bigquery.Table(table_ref, schema=schema)

            if partitioning_field:
                table.time_partitioning = bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY,
                    field=partitioning_field,  # name of column to use for partitioning
                )
            if clustering_fields:
                table.clustering_fields = clustering_fields
            if description:
                table.description = description

            # Make an API request.
            table = self.parent.client.create_table(table)

            print(f"Created a table '{table.table_id}'", end='')
            if table.time_partitioning.field:
                print(f", partitioned on a column '{table.time_partitioning.field}'")
            self.parent.dataset.update()

            return table

    class GA4:
        """utilities to manage GA4"""

        def __init__(self, parent):
            self.parent = parent
            self.clustering_fields = ['client_id', 'event_name']
            self.flat_table_id = 'flat'
            self.start_date = ''
            self.end_date = ''

        @property
        def first_date_recorded(self) -> str:
            """the first date GA4 data was transferred to BigQuery in YYYYMMDD format"""
            partitions = [t for t in self.parent.dataset.tables if t.startswith('events_')]
            if partitions:
                return sorted(partitions)[0].replace("events_", "")

        def set_dates(self, start_date: str, end_date: str) -> None:
            self.start_date = start_date.strip()
            self.end_date = end_date.strip()

        def audit_ep(self):
            sql = f'''--Event Parameterの発生状況
                SELECT
                    param.key AS parameter_name,
                    event_name,
                    DATE(timestamp_micros(event_timestamp), 'Asia/Tokyo') AS date,
                    COUNT(1) AS count,
                    COUNT(param.value.string_value) AS string,
                    COUNT(param.value.int_value) AS int,
                    COUNT(param.value.float_value) AS float,
                    COUNT(param.value.double_value) AS double,
                FROM
                    `analytics_{self.parent.parent.ga['4'].property.id}.events_*`
                        CROSS JOIN UNNEST(event_params) AS param
                WHERE
                    _TABLE_SUFFIX >= '{self.start_date}' AND _TABLE_SUFFIX <= '{self.end_date}'
                GROUP BY 1,2,3
                ORDER BY 1,2,3
            '''
            df = self.parent.run(sql, to_dataframe=True)
            return df

        def template_schema(self) -> list:
            self.parent.parent.launch_gs(constants.GOOGLE_SHEET_GA4_TEMPLATE_URL)
            if self.parent.parent.gs.sheet.select('推奨BQ'):
                return [d for d in self.parent.parent.gs.sheet.data if d['Valid']]

        def get_flat_schema(self, ep: list, up: list) -> list:
            gs_data = self.template_schema()
            basic_schema = [
                {
                    'name': d['Field Name'],
                    'type': d['Type'],
                    'description': d['Description']
                 } for d in gs_data
            ]

            custom_schema = []
            for d in up + ep:
                custom_schema.append({
                    'name': d['field_name'],
                    'type': ga4.convert_ga4_type_to_bq_type(d['type']),
                    'description': d['description']
                })
            return basic_schema + custom_schema

        def get_bq_schema(self, schema: list) -> List[bigquery.SchemaField]:
            """Converts a dictionary to list of bigquery.SchemaField
            for use with bigquery client library.
            Dict must contain name and type keys.
            """
            return [
                bigquery.SchemaField(
                    name=x['name'],
                    field_type=x['type'],
                    description=x.get('description') if x.get('description') else '')
                for x in schema
            ]

        def create_flat_table(self, schema: list) -> None:
            """Create a table to store flattened GA4 data."""
            print(f"Creating a table to store flattened GA4 data.")
            # Make an API request.
            self.parent.table.create(
                table_id=self.flat_table_id,
                description='This is a table to store flattened and optimized GA4 data exported.',
                schema=self.get_bq_schema(schema),
                partitioning_field='date',
                clustering_fields=self.clustering_fields
            )

        def flatten_events(
                self,
                date1: str,
                date2: str,
                event_parameters: list = None,
                user_properties: list = None,
                to: str = 'dataframe'
        ) -> None:
            """Flattened event tables exported from GA4"""
            if event_parameters is None:
                event_parameters = []
            if user_properties is None:
                user_properties = []

            sql = self.get_query_to_flatten_events(
                date1,
                date2,
                event_parameters=event_parameters,
                user_properties=user_properties,
            )

            if to == 'dataframe':
                # return the data as pandas dataframe
                df = self.parent.run(sql).to_dataframe()
                print(f"{len(df)} rows were retrieved.")
                return df
            elif to == 'table':
                # append the data to a table
                self.parent.table.select(self.flat_table_id)
                table_ref = self.parent.table.ref
                rows_before = self.parent.table.instance.num_rows
                print(f"The table '{self.flat_table_id}' had {rows_before} rows.")

                job_config = bigquery.QueryJobConfig(
                    # clustering_fields=self.clustering_fields,
                    destination=table_ref,
                    write_disposition=bigquery.WriteDisposition.WRITE_APPEND
                )
                # Start the query, passing in the extra configuration.
                query_job = self.parent.client.query(sql, job_config=job_config)  # Make an API request.
                result_iterator = query_job.result()  # Wait for the job to complete.

                rows_after = result_iterator.total_rows
                print(f"{rows_after - rows_before} rows were added for the period {date1} - {date2}")
                # return result_iterator
            else:
                print(f"Unknown destination: {to}")

        def get_query_to_flatten_events(
                self,
                date1: str,
                date2: str,
                event_parameters: list = None,
                user_properties: list = None,
                to: str = 'select'
        ):
            """Return a query to flatten GA4 event tables exported"""
            if event_parameters is None:
                event_parameters = []
            if user_properties is None:
                user_properties = []

            dataset = self.parent.dataset.id
            table_id = self.flat_table_id
            schema = self.template_schema()

            if to == 'scheduled_query':
                query = f'''--未処理のGA4生データを変換しflatテーブルへ追記
DECLARE last_exported_date DATE;
DECLARE next_date_to_be_processed DATE;
DECLARE processing_date STRING;

EXECUTE IMMEDIATE FORMAT("""
  --最後にエクスポートされたGA4データの日付
  SELECT CAST(RIGHT(table_name, 8) AS DATE FORMAT 'YYYYMMDD')
  FROM `{dataset}.INFORMATION_SCHEMA.TABLES`
  WHERE REGEXP_CONTAINS(table_name, r'^events_2')
  ORDER BY 1 DESC
  LIMIT 1
""") INTO last_exported_date;

EXECUTE IMMEDIATE FORMAT("""
  --未処理の最初の日付
  SELECT DATE_ADD(date, INTERVAL 1 DAY)
  FROM `{dataset}.{table_id}`
  WHERE date > DATE_SUB('%t', INTERVAL 10 DAY)--節約のため直近に絞る
  GROUP BY 1
  ORDER BY 1 DESC
  LIMIT 1
""", last_exported_date) INTO next_date_to_be_processed;

--処理すべき日付を判定
IF next_date_to_be_processed = last_exported_date THEN
  SET processing_date = FORMAT_DATE("%Y%m%d", last_exported_date);
  SELECT FORMAT('New data for %t found. Processing events_%s...', next_date_to_be_processed, processing_date);
ELSEIF next_date_to_be_processed < last_exported_date THEN
  SET processing_date = FORMAT_DATE("%Y%m%d", next_date_to_be_processed);
  SELECT FORMAT('New data for %t found, but processing events_%s first...', last_exported_date, processing_date);
ELSE
  --skip
  RAISE USING MESSAGE = FORMAT("%t has already been processed. Skipping...", last_exported_date);
END IF;

--go ahead
INSERT INTO `{dataset}.{table_id}` (
  '''

            else:
                query = ''
            query += f'''--Flatten GA4 events
  SELECT'''

            for s in schema:
                if s['Category']:
                    query += f'''
    --{s['Category']}'''
                query += f'''
    {s['Select']} AS {s['Field Name']},'''

            if user_properties:
                query += f'''
    --Custom User Properties'''
                for d in user_properties:
                    query += f'''
    (SELECT value.{d['type']}_value FROM UNNEST(user_properties) WHERE key = '{d['key']}') AS {d['field_name']},'''

            if event_parameters:
                query += f'''
    --Custom Event Parameters'''
                for d in event_parameters:
                    query += f'''
    (SELECT value.{d['type']}_value FROM UNNEST(event_params) WHERE key = '{d['key']}') AS {d['field_name']},'''

            query += f'''
  FROM
    `{dataset}.events_*`
  WHERE
    '''

            if to == 'scheduled_query':
                query += f"""_TABLE_SUFFIX = processing_date
);

SELECT FORMAT("%d rows (%d bytes) of data for %s were successfully inserted.", @@row_count, @@script.bytes_processed, processing_date);
"""

            else:
                query += f"""_TABLE_SUFFIX >= '{date1}' AND _TABLE_SUFFIX <= '{date2}'"""

            return query

        def schedule_query_to_flatten_events(
                self,
                event_parameters: list = None,
                user_properties: list = None
        ):
            """Save a scheduled query to flatten event tables exported from GA4"""
            if event_parameters is None:
                event_parameters = []
            if user_properties is None:
                user_properties = []
            sql = self.get_query_to_flatten_events('', '',
                                                   event_parameters,
                                                   user_properties,
                                                   to='scheduled_query')

            dataset_id = self.parent.dataset.id
            table_id = self.parent.table.id

            transfer_config = bigquery_datatransfer.TransferConfig(
                # destination_dataset_id=dataset_id,
                display_name=f"Flatten GA4 events for {dataset_id}",
                data_source_id="scheduled_query",
                params={
                    "query": sql,
                },
                schedule="every 4 hours",
            )

            request = bigquery_datatransfer.CreateTransferConfigRequest(
                parent=self.parent.dts_client.common_location_path(
                    self.parent.id,
                    self.parent.dataset.instance.location),
                transfer_config=transfer_config,
            )

            try:
                response = self.parent.dts_client.create_transfer_config(request=request)
                print(f"Scheduled query was created: {response.name}")
                return response
            except PermissionDenied as e:
                print("APIを実行する権限がありません。")
                m = re.search(r'reason: "([^"]+)', str(sys.exc_info()[1]))
                if m:
                    reason = m.group(1)
                    if reason == 'SERVICE_DISABLED':
                        print("GCPのプロジェクトでBigQuery Data Transfer APIを有効化してください。")
                message = getattr(e, 'message', repr(e))
                print(message)



"""
Functions for handling Authentications
"""

import json
import logging
import os
from collections import defaultdict

import google.oauth2.credentials
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow

_REQUIRED_CONFIG_KEYS = frozenset(("auth_uri", "token_uri", "client_id"))
LOGGER = logging.getLogger(__name__)


def _is_service_account(json_text: str):
    """Return true if the provided text is a JSON service credentials file."""
    try:
        key_obj = json.loads(json_text)
    except json.JSONDecodeError:
        return False
    if not key_obj or key_obj.get('type', '') != 'service_account':
        return False
    return True


def _is_service_account_json(json_path: str):
    """Return true if the provided JSON file is for a service account."""
    with open(json_path, 'r') as f:
        return _is_service_account(f.read())


def get_credential_type(client_config: dict):
    """Gets a client type from client configuration loaded from a Google-format client secrets file.

    Args:
        client_config (Mapping[str, Any]): The client
            configuration in the Google `client secrets`_ format.

    Returns:
        client_type [str]: The client type, either ``'service_account'`` or ``'web'`` or ``'installed'``
    """
    if client_config.get('type', '') == "service_account":
        return "service_account"
    elif "web" in client_config:
        client_type = "web"
    elif "installed" in client_config:
        client_type = "installed"
    else:
        return None
    config = client_config[client_type]
    if _REQUIRED_CONFIG_KEYS.issubset(config.keys()):
        return client_type


def get_credential_type_from_file(json_path: str):
    """Gets a client type from a Google client secrets file.

        Args:
            json_path (str): The path to the client secrets .json file.

        Returns:
            client_type [str]: The client type, either ``'service_account'`` or ``'web'`` or ``'installed'``
        """
    try:
        with open(json_path, "r") as fp:
            client_config = json.load(fp)
    except (OSError, json.JSONDecodeError):
        LOGGER.debug("Skipping non JSON credential file: %s", json_path)
        return None

    if not isinstance(client_config, dict):
        LOGGER.debug('Credential file %s does not contain a JSON object; skipping.', json_path)
        return None

    return get_credential_type(client_config)


def get_credential_type_from_info(info: dict) -> str:
    if isinstance(info, dict):
        if info.get("type") == "service_account":
            return "service_account"
        if "installed" in info:
            return "installed"
        if "web" in info:
            return "web"
    return "unknown"


def get_json_files_from_dir(json_dir: str):
    """Gets a list of valid credentials json files from a directory recursively"""
    json_files = defaultdict(lambda: {})
    for root, dirs, files in os.walk(json_dir):
        for file in files:
            if file.endswith('.json'):
                client_type = get_credential_type_from_file(os.path.join(root, file))
                if client_type == 'service_account':
                    json_files['Service Account'][file] = os.path.join(root, file)
                elif client_type in ['installed', 'web']:
                    json_files['OAuth'][file] = os.path.join(root, file)
    return json_files


def get_cache_path(json_path: str):
    """Gets the path to the Google user credentials based on the provided source file
    """
    dir_path = os.path.join(os.path.expanduser("~"), ".config")
    os.makedirs(dir_path, exist_ok=True)
    base_name = os.path.splitext(os.path.basename(json_path))[0]
    return os.path.join(dir_path, f"cache_{base_name}.json")


def save_credentials(file_path: str, credentials: Credentials):
    """Save Credentials to cache file
    """
    cache_path = get_cache_path(file_path)
    with open(cache_path, 'w') as w:
        LOGGER.debug(f"saving credentials to {cache_path}")
        w.write(credentials.to_json())
    return credentials


def load_credentials(file_path: str, scopes: list):
    """Load Credentials from cache file
    """
    cache_path = get_cache_path(file_path)
    if os.path.isfile(cache_path):
        LOGGER.debug(f"loading credentials from {cache_path}")
        return Credentials.from_authorized_user_file(cache_path, scopes=scopes)


def load_service_account_credentials_from_info(info: dict, scopes: list):
    if not isinstance(info, dict) or info.get("type") != "service_account":
        raise ValueError("service_account info required")
    credentials = service_account.Credentials.from_service_account_info(info, scopes=scopes)
    if not credentials.valid:
        request = google.auth.transport.requests.Request()
        try:
            credentials.refresh(request)
        except google.auth.exceptions.RefreshError as exc:
            email = info.get("client_email") or getattr(credentials, "service_account_email", None)
            if email:
                message = f"指定の {email} のサービスアカウントは存在しない、または無効です。"
            else:
                message = "指定したサービスアカウントは存在しない、または無効です。"
            LOGGER.error(message)
            LOGGER.debug(f"Service account refresh error detail: {exc}")
            return None
    return credentials


def delete_credentials(cache_file: str = "creden-cache.json"):
    """Delete Credentials cache file
    """
    if os.path.isfile(cache_file):
        LOGGER.debug(f"deleting cache file {cache_file}")
        os.remove(cache_file)


def get_oauth_redirect(client_secret_file: str, scopes: list):
    """Run OAuth2 Flow"""
    flow = InstalledAppFlow.from_client_secrets_file(
        client_secret_file,
        scopes=scopes,
        redirect_uri="urn:ietf:wg:oauth:2.0:oob"
    )
    auth_url, _ = flow.authorization_url(prompt="consent")
    return flow, auth_url


def get_oauth_redirect_from_info(info: dict, scopes: list):
    # mirrors get_oauth_redirect(file, scopes) but uses in-memory client config
    flow = InstalledAppFlow.from_client_config(info, scopes=scopes)
    flow.redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
    auth_url, _ = flow.authorization_url(
        prompt='consent',
        access_type='offline',
        include_granted_scopes='true'
    )
    return flow, auth_url


def get_token(flow, code: str):
    flow.fetch_token(code=code)
    return flow.credentials


def load_service_account_credentials_from_file(path: str, scopes: list):
    """Gets service account credentials from JSON file at ``path``.

    :param path: Path to credentials JSON file.
    :param scopes: A list of scopes to use when authenticating to Google APIs.
    :return: google.oauth2.service_account.Credentials
    """
    credentials = service_account.Credentials.from_service_account_file(path, scopes=scopes)
    if not credentials.valid:
        request = google.auth.transport.requests.Request()
        try:
            credentials.refresh(request)
        except google.auth.exceptions.RefreshError as exc:
            email = getattr(credentials, "service_account_email", None)
            if email:
                message = f"指定の {email} のサービスアカウントは存在しない、または無効です。"
            else:
                message = "指定したサービスアカウントは存在しない、または無効です。"
            LOGGER.error(message)
            LOGGER.debug("Service account refresh error detail: %s", exc)
            return None
    return credentials


__all__ = [
    "_is_service_account",
    "_is_service_account_json",
    "get_credential_type",
    "get_credential_type_from_file",
    "get_credential_type_from_info",
    "get_json_files_from_dir",
    "get_cache_path",
    "save_credentials",
    "load_credentials",
    "load_service_account_credentials_from_info",
    "delete_credentials",
    "get_oauth_redirect",
    "get_oauth_redirect_from_info",
    "get_token",
    "load_service_account_credentials_from_file",
]



from __future__ import annotations

import re
from typing import Optional

import pandas as pd


def convert_filter_to_event_scope(filter_d: Optional[str]) -> Optional[str]:
    """session系フィルタディメンションをevent系に変換
    
    GA4 APIでは、session系ディメンション（sessionDefaultChannelGroupなど）と
    event系ディメンション（defaultChannelGroupなど）でフィルタの互換性がない。
    この関数は、session系ディメンションを使ったfilter_dをevent系クエリで
    使用できるように変換する。
    
    ディメンション名のみを変換し、値や正規表現パターンは変換しない。
    
    Args:
        filter_d: フィルタ文字列（例: "sessionDefaultChannelGroup==Organic Social"）
    
    Returns:
        event系に変換されたフィルタ文字列（例: "defaultChannelGroup==Organic Social"）
        Noneまたは空文字列が渡された場合はそのまま返す
    
    Examples:
        >>> convert_filter_to_event_scope("sessionDefaultChannelGroup==Organic Social")
        'defaultChannelGroup==Organic Social'
        >>> convert_filter_to_event_scope("sessionMedium==social;sessionSource==facebook")
        'medium==social;source==facebook'
        >>> convert_filter_to_event_scope("page=~sessionMedium")
        'page=~sessionMedium'  # 値側は変換されない
    """
    if not filter_d:
        return filter_d
    
    # session系 → event系ディメンション変換
    conversions = {
        'sessionDefaultChannelGroup': 'defaultChannelGroup',
        'sessionSourceMedium': 'sourceMedium',
        'sessionMedium': 'medium',
        'sessionSource': 'source',
        'sessionCampaignId': 'campaignId',
        'sessionCampaignName': 'campaignName',
        'sessionManualTerm': 'manualTerm',
        'sessionManualSource': 'manualSource',
        'sessionManualMedium': 'manualMedium',
        'sessionManualSourceMedium': 'manualSourceMedium',
        'sessionManualCampaignId': 'manualCampaignId',
        'sessionManualCampaignName': 'manualCampaignName',
        'sessionManualAdContent': 'manualAdContent',
    }
    
    # セミコロン区切りで分割
    parts = filter_d.split(';')
    converted_parts = []

    for part in parts:
        part = part.strip()
        # 演算子で分割 (==, =~, =@, !=, !~, !@)
        match = re.match(r'^\s*([^=!]+?)\s*(==|=~|=@|!=|!~|!@)\s*(.*)$', part)
        if match:
            dim, op, value = match.groups()
            # ディメンション名のみ変換（値は変換しない）
            dim_converted = conversions.get(dim.strip(), dim.strip())
            converted_parts.append(f"{dim_converted}{op}{value}")
        else:
            # 演算子が見つからない場合はそのまま
            converted_parts.append(part)
    
    return ';'.join(converted_parts)


def classify_source_channel(
    df: pd.DataFrame,
    channel_col: str = "channel",
    medium_col: str = "medium",
    source_col: str = "source",
    custom_channels=None,
) -> pd.DataFrame:
    """sourceとchannelを分類する（コア関数）
    
    megatonの既存ロジックをベースに、sourceとchannelの両方を返す版。
    AI判定は正規表現を使った網羅的なパターンマッチングを使用。
    
    Args:
        df: データフレーム
        channel_col: チャネル列名（default: "channel"）
        medium_col: メディア列名（default: "medium"）
        source_col: ソース列名（default: "source"）
        custom_channels: カスタムチャネル定義（default: None）
            {
                "チャネル名": {
                    "normalize": {"表示名": r"パターン", ...},
                    "detect": [r"パターン", ...]
                }
            }
            または簡易形式（正規表現リスト）:
            {
                "チャネル名": [r"パターン", ...]  # detectのみ、正規表現として扱われる
            }
        
    Returns:
        pd.DataFrame: 2列のDataFrame。列名はsource_col, channel_colパラメータに従う
        
    Examples:
        >>> result = classify_source_channel(df)
        >>> df[["source", "channel"]] = result
        
        # 非デフォルト列名
        >>> result = classify_source_channel(df, source_col="my_src", channel_col="my_ch")
        >>> df[["my_src", "my_ch"]] = result
        
        # Group（project_x用）- 簡易形式は正規表現
        >>> result = classify_source_channel(
        ...     df,
        ...     custom_channels={"Group": [r"dentamap\.jp", r"haisha-yoyaku\.jp"]}
        ... )
        
        # client_x Internal（dei用）
        >>> result = classify_source_channel(
        ...     df,
        ...     custom_channels={
        ...         "client_x Internal": [
        ...             "extra\\.client_x\\.co\\.jp",
        ...             "sharepoint",
        ...             "teams",
        ...         ]
        ...     }
        ... )
    """
    for col in [channel_col, source_col]:
        if col not in df.columns:
            raise ValueError(f"Missing column: {col}")
    
    # ======================
    # custom_channelsの正規化（簡易形式→完全形式）
    # ======================
    normalized_custom = {}
    if custom_channels:
        for channel_name, config in custom_channels.items():
            if isinstance(config, (list, tuple)):
                # 簡易形式: [patterns...] → {"detect": [patterns...]}
                normalized_custom[channel_name] = {
                    "normalize": {},
                    "detect": list(config),
                }
            elif isinstance(config, dict):
                # 完全形式
                normalized_custom[channel_name] = {
                    "normalize": config.get("normalize", {}),
                    "detect": config.get("detect", []),
                }
    
    # ======================
    # チャネル別パターン定義
    # ======================
    
    channel_patterns = {
        "AI": {
            "normalize": {
                "ChatGPT": r"chatgpt|chat\.openai\.com|openai\.com",
                "Copilot": r"copilot|bing\.com/chat|microsoftcopilot",
                "Gemini": r"gemini|bard|aistudio\.google\.com|makersuite\.google\.com",
                "Claude": r"claude|anthropic\.com",
                "Perplexity": r"perplexity|pplx\.ai",
            },
            "detect": [],
        },
        
        "Organic Search": {
            "normalize": {
                "docomo.ne.jp": r"service\.smt\.docomo\.ne\.jp|docomo\.ne\.jp",
                "bing": r"(?:^|\.)bing\.com$",
                "auone.jp": r"(?:^|\.)auone\.jp$",
            },
            "detect": [
                r"docomo\.ne\.jp",
                r"\.jword\.jp",
                r"jp\.hao123\.com",
                r"\bsearch\b",
            ],
        },
        
        "Organic Social": {
            "normalize": {
                "Facebook": r"facebook(\.com)?|fb\.com",
                "X": r"^t\.co$|twitter\.com|x\.com",
                "Instagram": r"^ig$|instagram\.com|\binstagram\b",
                "YouTube": r"youtube(\.com)?|youtu\.be",
                "TikTok": r"tiktok(\.com)?",
                "Threads": r"threads(\.com|\.net)?",
            },
            "detect": [],
        },
    }
    
    # カスタムチャネルをマージ
    channel_patterns.update(normalized_custom)
    
    # ======================
    # コンパイル
    # ======================
    
    # 各channelの判定用パターン（正規化 + 追加判定を統合）
    channel_detect_patterns = {}
    for channel_name, config in channel_patterns.items():
        all_patterns = list(config["normalize"].values()) + config["detect"]
        if all_patterns:
            channel_detect_patterns[channel_name] = re.compile(
                "|".join(all_patterns),
                re.IGNORECASE
            )
    
    # Source正規化用パターン（全channelの正規化を統合）
    source_normalizations = []
    for config in channel_patterns.values():
        source_normalizations.extend([
            (pattern, name) for name, pattern in config["normalize"].items()
        ])
    source_normalize_compiled = [
        (re.compile(pattern, re.IGNORECASE), replacement)
        for pattern, replacement in source_normalizations
    ]
    
    # 正規化済みSNS名のセット（高速判定用）
    normalized_sns_names = set(channel_patterns["Organic Social"]["normalize"].keys())
    
    # ======================
    # 判定関数
    # ======================
    
    def _normalize_source(src: str) -> str:
        """source文字列を正規化"""
        for pattern, replacement in source_normalize_compiled:
            if pattern.search(src):
                return replacement
        return src
    
    def _classify_row(row):
        """1行ごとに(source, channel)のタプルを返す"""
        channel = str(row.get(channel_col, ""))
        medium = str(row.get(medium_col, "")).lower()
        source_raw = str(row.get(source_col, ""))
        source = source_raw.lower().replace("www.", "")
        
        # 先に正規化
        normalized = _normalize_source(source_raw)
        
        # AI判定
        if channel_detect_patterns.get("AI") and (
            channel_detect_patterns["AI"].search(source) or 
            channel_detect_patterns["AI"].search(medium)
        ):
            return normalized, "AI"
        
        # Map判定（既存ロジック）
        if medium == "map" or "maps." in source or ".maps." in source:
            return normalized, "Map"
        
        # Referralの再分類
        if channel == "Referral":
            # Organic Search判定
            if channel_detect_patterns.get("Organic Search") and \
               channel_detect_patterns["Organic Search"].search(source):
                return normalized, "Organic Search"
            
            # Organic Social判定
            # 1. 正規化済みSNS名かチェック
            if normalized in normalized_sns_names:
                return normalized, "Organic Social"
            # 2. 追加のSNSパターン
            if channel_detect_patterns.get("Organic Social") and \
               channel_detect_patterns["Organic Social"].search(source):
                return normalized, "Organic Social"
            
            # カスタムチャネル判定（Groupなど）
            for channel_name, pattern in channel_detect_patterns.items():
                if channel_name not in ["AI", "Organic Search", "Organic Social"]:
                    if pattern.search(source):
                        return normalized, channel_name
        
        # fallback
        return normalized, channel
    
    result = df.apply(_classify_row, axis=1, result_type="expand")
    result.columns = [source_col, channel_col]
    return result


def classify_channel(
    df: pd.DataFrame,
    channel_col: str = "channel",
    medium_col: str = "medium",
    source_col: str = "source",
    custom_channels=None,
) -> pd.Series:
    """channelのみを分類（ラッパー関数）
    
    後方互換性のため、channel列のみを返すシンプル版。
    内部でclassify_source_channel()を呼び出す。
    
    Args:
        df: データフレーム
        channel_col: チャネル列名（default: "channel"）
        medium_col: メディア列名（default: "medium"）
        source_col: ソース列名（default: "source"）
        custom_channels: カスタムチャネル定義（default: None）
        
    Returns:
        pd.Series: channel列
        
    Examples:
        >>> df["channel"] = classify_channel(df)
        
        # Group指定（簡易形式：正規表現リスト）
        >>> df["channel"] = classify_channel(
        ...     df,
        ...     custom_channels={"Group": [r"dentamap\.jp", r"example\.com"]}
        ... )
    """
    result = classify_source_channel(
        df,
        channel_col=channel_col,
        medium_col=medium_col,
        source_col=source_col,
        custom_channels=custom_channels,
    )
    return result[channel_col]



import builtins
import importlib
import sys
import types

import pytest


def ensure_module(name: str):
    parts = name.split('.')
    module_name = ''
    parent = None
    for part in parts:
        module_name = f"{module_name}.{part}" if module_name else part
        module = sys.modules.get(module_name)
        if module is None:
            module = types.ModuleType(module_name)
            sys.modules[module_name] = module
            if parent is not None:
                setattr(parent, part, module)
        parent = module
    return module


def get_module(name: str):
    try:
        return importlib.import_module(name)
    except Exception:
        return ensure_module(name)


def _ensure_ga_stubs():
    data_mod = get_module("google.analytics.data")
    if not hasattr(data_mod, "BetaAnalyticsDataClient"):
        class _StubDataClient:
            pass

        data_mod.BetaAnalyticsDataClient = _StubDataClient

    admin_mod = get_module("google.analytics.admin")
    if not hasattr(admin_mod, "AnalyticsAdminServiceClient"):
        class _StubAdminClient:
            pass

        admin_mod.AnalyticsAdminServiceClient = _StubAdminClient


def _clear_modules(prefix: str):
    for name in list(sys.modules.keys()):
        if name == prefix or name.startswith(prefix + "."):
            del sys.modules[name]


def _block_ipywidgets(monkeypatch):
    real_import = builtins.__import__

    def fake_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "ipywidgets" or name.startswith("ipywidgets."):
            raise ModuleNotFoundError("No module named 'ipywidgets'")
        return real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", fake_import)


def test_import_start_without_ipywidgets(monkeypatch):
    _ensure_ga_stubs()
    _clear_modules("megaton")
    _clear_modules("ipywidgets")
    _block_ipywidgets(monkeypatch)

    start = importlib.import_module("megaton.start")
    assert start.Megaton


def test_widgets_raise_without_ipywidgets(monkeypatch):
    _ensure_ga_stubs()
    _clear_modules("megaton")
    _clear_modules("ipywidgets")
    _block_ipywidgets(monkeypatch)

    widgets = importlib.import_module("megaton.ui.widgets")
    with pytest.raises(widgets.WidgetsUnavailableError) as excinfo:
        widgets.dropdown_menu("Label", "Default")

    message = str(excinfo.value)
    assert "ipywidgets" in message
    assert "headless=True" in message



import os


def install():
    """Install packages for GA4."""
    os.system("pip install -U -q google-analytics-admin")
    os.system("pip install -U -q google-analytics-data")
    # update packages
    # os.system("pip install -U -q google-api-core")
    # os.system("pip install -U -q google-auth")



import json
from types import SimpleNamespace

import gspread
import pandas as pd
import pytest
import requests

from megaton import errors
from megaton.gsheet import MegatonGS


def _new_gs():
    gs = MegatonGS.__new__(MegatonGS)
    return gs


def _api_error(message: str, code: int = 403) -> gspread.exceptions.APIError:
    resp = requests.Response()
    resp.status_code = code
    resp._content = json.dumps(
        {"error": {"code": code, "message": message, "status": message}}
    ).encode()
    return gspread.exceptions.APIError(resp)


class _DummyClient:
    def __init__(self, result):
        self.result = result

    def open_by_url(self, _url):
        if isinstance(self.result, Exception):
            raise self.result
        return self.result


class _FakeCell:
    def __init__(self, row, value):
        self.row = row
        self.value = value


class _FakeWorksheet:
    def __init__(self, title="Sheet1", row_count=5, col_count=3):
        self.title = title
        self.id = 123
        self.row_count = row_count
        self.col_count = col_count
        self.records = []
        self.cleared = False
        self.added_rows = 0
        self.added_cols = 0
        self.batch_clears = []
        self.frozen = None
        self.updated = {}
        self.last_acell = None

    def clear(self):
        self.cleared = True

    def get_all_records(self):
        return list(self.records)

    def range(self, _r1, _c1, _r2, _c2):
        return [_FakeCell(1, "h"), _FakeCell(2, "v")]

    def add_rows(self, n):
        self.row_count += n
        self.added_rows += n

    def add_cols(self, n):
        self.col_count += n
        self.added_cols += n

    def batch_clear(self, ranges):
        self.batch_clears.extend(ranges)

    def freeze(self, rows=None, cols=None):
        self.frozen = (rows, cols)

    def acell(self, address):
        self.last_acell = address
        return SimpleNamespace(value=f"value:{address}")

    def update(self, address, value):
        self.updated[address] = value


class _FakeSpreadsheet:
    def __init__(self, worksheet, worksheet_error=None):
        self._worksheet = worksheet
        self._worksheet_error = worksheet_error
        self.batch_updates = []
        self.deleted = None

    def worksheet(self, _name):
        if self._worksheet_error is not None:
            raise self._worksheet_error
        return self._worksheet

    def add_worksheet(self, title, rows, cols):
        self._worksheet = _FakeWorksheet(title=title, row_count=rows, col_count=cols)

    def del_worksheet(self, ws):
        self.deleted = ws

    def batch_update(self, payload):
        self.batch_updates.append(payload)


def _build_sheet(worksheet=None, worksheet_error=None):
    ws = worksheet or _FakeWorksheet()
    parent = SimpleNamespace(
        _client=True,
        _driver=_FakeSpreadsheet(ws, worksheet_error=worksheet_error),
    )
    sheet = MegatonGS.Sheet(parent)
    sheet._driver = ws
    return sheet, ws


def test_resolve_timeout_invalid_env_falls_back(monkeypatch):
    monkeypatch.setenv("MEGATON_GS_TIMEOUT", "invalid")
    gs = _new_gs()
    assert gs._resolve_timeout(None) == 180.0


def test_open_raises_bad_url_format():
    gs = _new_gs()
    gs._client = _DummyClient(gspread.exceptions.NoValidUrlKeyFound())

    with pytest.raises(errors.BadUrlFormat):
        gs.open("bad-url")


def test_open_raises_bad_permission_from_permission_error():
    gs = _new_gs()
    gs._client = _DummyClient(PermissionError("no permission"))

    with pytest.raises(errors.BadPermission):
        gs.open("https://docs.google.com/spreadsheets/d/xxxx")


def test_open_maps_api_errors():
    for message, expected in [
        ("disabled", errors.ApiDisabled),
        ("PERMISSION_DENIED", errors.BadPermission),
        ("NOT_FOUND", errors.UrlNotFound),
    ]:
        gs = _new_gs()

        class _Driver:
            @property
            def title(self):
                raise _api_error(message)

        gs._client = _DummyClient(_Driver())
        with pytest.raises(expected):
            gs.open("https://docs.google.com/spreadsheets/d/xxxx")


def test_open_success_with_sheet_selection():
    gs = _new_gs()
    selected = []
    driver = SimpleNamespace(title="Book", url="u")
    gs._client = _DummyClient(driver)
    gs.sheet = SimpleNamespace(select=lambda name: selected.append(name))

    title = gs.open("https://docs.google.com/spreadsheets/d/xxxx", sheet="Report")

    assert title == "Book"
    assert selected == ["Report"]


def test_sheet_select_raises_sheet_not_found():
    sheet, _ = _build_sheet(worksheet_error=gspread.exceptions.WorksheetNotFound("missing"))

    with pytest.raises(errors.SheetNotFound):
        sheet.select("missing")


def test_sheet_select_maps_api_error():
    for message, expected in [
        ("disabled", errors.ApiDisabled),
        ("PERMISSION_DENIED", errors.BadPermission),
    ]:
        sheet, _ = _build_sheet(worksheet_error=_api_error(message))
        with pytest.raises(expected):
            sheet.select("x")


def test_save_data_mode_w_maps_clear_errors(monkeypatch):
    sheet, ws = _build_sheet()
    df = pd.DataFrame([{"a": 1}])

    def _raise():
        raise _api_error("PERMISSION_DENIED")

    monkeypatch.setattr(sheet, "clear", _raise)
    with pytest.raises(errors.BadPermission):
        sheet.save_data(df, mode="w")

    def _raise2():
        raise _api_error("disabled")

    monkeypatch.setattr(sheet, "clear", _raise2)
    with pytest.raises(errors.ApiDisabled):
        sheet.save_data(df, mode="w")

    assert ws.cleared is False


def test_save_data_mode_a_adds_rows_and_writes(monkeypatch):
    sheet, ws = _build_sheet(worksheet=_FakeWorksheet(row_count=2, col_count=2))
    df = pd.DataFrame([{"a": 1}, {"a": 2}])
    calls = {}

    monkeypatch.setattr(sheet, "_refresh", lambda: None)

    def _record(*args, **kwargs):
        calls["args"] = args
        calls["kwargs"] = kwargs

    monkeypatch.setitem(sheet.save_data.__globals__, "set_with_dataframe", _record)
    result = sheet.save_data(df, mode="a", include_index=False)

    assert result is True
    assert ws.added_rows > 0
    assert calls["kwargs"]["include_column_header"] is False
    assert calls["kwargs"]["resize"] is False


def test_overwrite_data_from_row_delegates_when_row_le_1(monkeypatch):
    sheet, _ = _build_sheet()
    df = pd.DataFrame([{"a": 1}])

    monkeypatch.setattr(sheet, "overwrite_data", lambda _df, include_index=False: "ok")
    assert sheet.overwrite_data_from_row(df, row=1) == "ok"


def test_overwrite_data_from_row_maps_batch_clear_errors(monkeypatch):
    sheet, ws = _build_sheet()
    df = pd.DataFrame([{"a": 1}])

    def _raise(_ranges):
        raise _api_error("disabled")

    monkeypatch.setattr(ws, "batch_clear", _raise)
    with pytest.raises(errors.ApiDisabled):
        sheet.overwrite_data_from_row(df, row=2)


def test_overwrite_data_from_row_resizes_and_writes(monkeypatch):
    sheet, ws = _build_sheet(worksheet=_FakeWorksheet(row_count=2, col_count=1))
    df = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    calls = {}

    monkeypatch.setattr(sheet, "_refresh", lambda: None)

    def _record(*args, **kwargs):
        calls["kwargs"] = kwargs

    monkeypatch.setitem(
        sheet.overwrite_data_from_row.__globals__,
        "set_with_dataframe",
        _record,
    )
    result = sheet.overwrite_data_from_row(df, row=3)

    assert result is True
    assert "3:2" in ws.batch_clears
    assert ws.added_rows > 0
    assert ws.added_cols > 0
    assert calls["kwargs"]["row"] == 3
    assert calls["kwargs"]["include_column_header"] is True


def test_cell_select_get_and_set_data():
    sheet, ws = _build_sheet()
    cell = MegatonGS.Sheet.Cell(sheet)

    value = cell.select(2, 3)
    assert value == "value:C2"

    cell.select("B2")
    cell.data = "x"
    assert ws.updated["B2"] == "x"
